from typing import Dict, List, Optional

from pydantic import Field
from typing_extensions import TypedDict

from .base import GuardrailConfigModel


class JavelinGuardInput(TypedDict):
    text: str


class JavelinGuardRequest(TypedDict):
    input: JavelinGuardInput
    config: Optional[Dict]
    metadata: Optional[Dict]


class JavelinPromptInjectionCategories(TypedDict):
    prompt_injection: bool
    jailbreak: bool


class JavelinPromptInjectionCategoryScores(TypedDict):
    prompt_injection: float
    jailbreak: float


class JavelinPromptInjectionResults(TypedDict):
    categories: JavelinPromptInjectionCategories
    category_scores: JavelinPromptInjectionCategoryScores
    reject_prompt: str


class JavelinPromptInjectionAssessment(TypedDict):
    results: JavelinPromptInjectionResults
    request_reject: bool


class JavelinTrustSafetyCategories(TypedDict):
    violence: bool
    weapons: bool
    hate_speech: bool
    crime: bool
    sexual: bool
    profanity: bool


class JavelinTrustSafetyCategoryScores(TypedDict):
    violence: float
    weapons: float
    hate_speech: float
    crime: float
    sexual: float
    profanity: float


class JavelinTrustSafetyResults(TypedDict):
    categories: JavelinTrustSafetyCategories
    category_scores: JavelinTrustSafetyCategoryScores


class JavelinTrustSafetyAssessment(TypedDict):
    results: JavelinTrustSafetyResults
    request_reject: bool


class JavelinLanguageDetectionResults(TypedDict):
    lang: str
    prob: float


class JavelinLanguageDetectionAssessment(TypedDict):
    results: JavelinLanguageDetectionResults
    request_reject: bool


class JavelinGuardResponse(TypedDict):
    assessments: List[
        Dict[
            str,
            JavelinPromptInjectionAssessment
            | JavelinTrustSafetyAssessment
            | JavelinLanguageDetectionAssessment,
        ]
    ]


class JavelinGuardrailConfigModel(GuardrailConfigModel):
    """Configuration parameters for the Javelin guardrail"""

    guard_name: Optional[str] = Field(
        default=None, description="Name of the Javelin guard to use"
    )
    api_version: Optional[str] = Field(
        default="v1", description="API version for Javelin service"
    )
    metadata: Optional[Dict] = Field(
        default=None, description="Additional metadata to send with requests"
    )
    application: Optional[str] = Field(
        default=None, description="Application name for Javelin service"
    )
    config: Optional[Dict] = Field(
        default=None, description="Configuration parameters for Javelin service"
    )

    @staticmethod
    def ui_friendly_name() -> str:
        return "Javelin Guardrails"



"""
Common utilities for Azure AI OCR providers.

This module provides routing logic to determine which OCR configuration to use
based on the model name.
"""

from typing import TYPE_CHECKING, Optional

from litellm._logging import verbose_logger

if TYPE_CHECKING:
    from litellm.llms.base_llm.ocr.transformation import BaseOCRConfig


def get_azure_ai_ocr_config(model: str) -> Optional["BaseOCRConfig"]:
    """
    Determine which Azure AI OCR configuration to use based on the model name.
    
    Azure AI supports multiple OCR services:
    - Azure Document Intelligence: azure_ai/doc-intelligence/<model>
    - Mistral OCR (via Azure AI): azure_ai/<model>
    
    Args:
        model: The model name (e.g., "azure_ai/doc-intelligence/prebuilt-read", 
               "azure_ai/pixtral-12b-2409")
    
    Returns:
        OCR configuration instance for the specified model
        
    Examples:
        >>> get_azure_ai_ocr_config("azure_ai/doc-intelligence/prebuilt-read")
        <AzureDocumentIntelligenceOCRConfig object>
        
        >>> get_azure_ai_ocr_config("azure_ai/pixtral-12b-2409")
        <AzureAIOCRConfig object>
    """
    from litellm.llms.azure_ai.ocr.document_intelligence.transformation import (
        AzureDocumentIntelligenceOCRConfig,
    )
    from litellm.llms.azure_ai.ocr.transformation import AzureAIOCRConfig

    # Check for Azure Document Intelligence models
    if "doc-intelligence" in model or "documentintelligence" in model:
        verbose_logger.debug(
            f"Routing {model} to Azure Document Intelligence OCR config"
        )
        return AzureDocumentIntelligenceOCRConfig()
    
    # Default to Mistral-based OCR for other azure_ai models
    verbose_logger.debug(f"Routing {model} to Azure AI (Mistral) OCR config")
    return AzureAIOCRConfig()




"""
Dynamic rate limiter v3 - Saturation-aware priority-based rate limiting
"""

import os
from datetime import datetime
from typing import TYPE_CHECKING, Callable, Dict, List, Optional, Union

from fastapi import HTTPException

import litellm
from litellm import ModelResponse, Router
from litellm._logging import verbose_proxy_logger
from litellm.caching.caching import DualCache
from litellm.integrations.custom_logger import CustomLogger
from litellm.proxy._types import UserAPIKeyAuth
from litellm.proxy.hooks.parallel_request_limiter_v3 import (
    RateLimitDescriptor,
    RateLimitDescriptorRateLimitObject,
    _PROXY_MaxParallelRequestsHandler_v3,
)
from litellm.proxy.hooks.rate_limiter_utils import convert_priority_to_percent
from litellm.proxy.utils import InternalUsageCache
from litellm.types.router import ModelGroupInfo
from litellm.types.utils import CallTypesLiteral

if TYPE_CHECKING:
    from litellm.types.utils import PriorityReservationSettings


def _get_priority_settings() -> "PriorityReservationSettings":
    """
    Get the priority reservation settings, guaranteed to be non-None.

    The settings are lazy-loaded in litellm.__init__ and always return an instance.
    This helper provides proper type narrowing for mypy.
    """
    settings = litellm.priority_reservation_settings
    if settings is None:
        # This should never happen due to lazy loading, but satisfy mypy
        from litellm.types.utils import PriorityReservationSettings

        return PriorityReservationSettings()
    return settings


class _PROXY_DynamicRateLimitHandlerV3(CustomLogger):
    """
    Saturation-aware priority-based rate limiter using v3 infrastructure.

    Key features:
    1. Model capacity ALWAYS enforced at 100% (prevents over-allocation)
    2. Priority usage tracked from first request (accurate accounting)
    3. Priority limits only enforced when saturated >= threshold
    4. Three-phase checking prevents partial counter increments
    5. Reuses v3 limiter's Redis-based tracking (multi-instance safe)

    How it works:
    - Phase 1: Read-only check of ALL limits (no increments)
    - Phase 2: Decide enforcement based on saturation
    - Phase 3: Increment counters only if request allowed
    - When under-saturated: priorities can borrow unused capacity (generous)
    - When saturated: strict priority-based limits enforced (fair)
    - Uses v3 limiter's atomic Lua scripts for race-free increments
    """

    def __init__(
        self,
        internal_usage_cache: DualCache,
        time_provider: Optional[Callable[[], datetime]] = None,
    ):
        self.internal_usage_cache = InternalUsageCache(dual_cache=internal_usage_cache)
        self.v3_limiter = _PROXY_MaxParallelRequestsHandler_v3(
            self.internal_usage_cache, time_provider=time_provider
        )

    def update_variables(self, llm_router: Router):
        self.llm_router = llm_router

    def _get_saturation_check_cache_ttl(self) -> int:
        """Get the configurable TTL for local cache when reading saturation values."""
        return _get_priority_settings().saturation_check_cache_ttl

    async def _get_saturation_value_from_cache(
        self,
        counter_key: str,
    ) -> Optional[str]:
        """
        Get saturation value with configurable local cache TTL.

        Uses DualCache with configurable TTL for local cache storage.
        TTL is configurable via litellm.priority_reservation_settings.saturation_check_cache_ttl

        Args:
            counter_key: The cache key for the saturation counter

        Returns:
            Counter value as string, or None if not found
        """
        local_cache_ttl = self._get_saturation_check_cache_ttl()

        return await self.internal_usage_cache.async_get_cache(
            key=counter_key,
            litellm_parent_otel_span=None,
            local_only=False,
            ttl=local_cache_ttl,
        )

    def _get_priority_weight(
        self, priority: Optional[str], model_info: Optional[ModelGroupInfo] = None
    ) -> float:
        """Get the weight for a given priority from litellm.priority_reservation"""
        weight: float = _get_priority_settings().default_priority
        if (
            litellm.priority_reservation is None
            or priority not in litellm.priority_reservation
        ):
            verbose_proxy_logger.debug(
                "Priority Reservation not set for the given priority."
            )
        elif priority is not None and litellm.priority_reservation is not None:
            if os.getenv("LITELLM_LICENSE", None) is None:
                verbose_proxy_logger.error(
                    "PREMIUM FEATURE: Reserving tpm/rpm by priority is a premium feature. Please add a 'LITELLM_LICENSE' to your .env to enable this.\nGet a license: https://docs.litellm.ai/docs/proxy/enterprise."
                )
            else:
                value = litellm.priority_reservation[priority]
                weight = convert_priority_to_percent(value, model_info)
        return weight

    def _get_priority_from_user_api_key_dict(
        self, user_api_key_dict: UserAPIKeyAuth
    ) -> Optional[str]:
        """
        Get priority from user_api_key_dict.

        Checks team metadata first (takes precedence), then falls back to key metadata.

        Args:
            user_api_key_dict: User authentication info

        Returns:
            Priority string if found, None otherwise
        """
        priority: Optional[str] = None

        # Check team metadata first (takes precedence)
        if user_api_key_dict.team_metadata is not None:
            priority = user_api_key_dict.team_metadata.get("priority", None)

        # Fall back to key metadata
        if priority is None:
            priority = user_api_key_dict.metadata.get("priority", None)

        return priority

    def _normalize_priority_weights(
        self, model_info: ModelGroupInfo
    ) -> Dict[str, float]:
        """
        Normalize priority weights if they sum to > 1.0

        Handles over-allocation: {key_a: 0.60, key_b: 0.80} -> {key_a: 0.43, key_b: 0.57}
        Converts absolute rpm/tpm values to percentages based on model capacity.
        """
        if litellm.priority_reservation is None:
            return {}

        # Convert all values to percentages first
        weights: Dict[str, float] = {}
        for k, v in litellm.priority_reservation.items():
            weights[k] = convert_priority_to_percent(v, model_info)

        total_weight = sum(weights.values())

        if total_weight > 1.0:
            normalized = {k: v / total_weight for k, v in weights.items()}
            verbose_proxy_logger.debug(
                f"Normalized over-allocated priorities: {weights} -> {normalized}"
            )
            return normalized

        return weights

    def _get_priority_allocation(
        self,
        model: str,
        priority: Optional[str],
        normalized_weights: Dict[str, float],
        model_info: Optional[ModelGroupInfo] = None,
    ) -> tuple[float, str]:
        """
        Get priority weight and pool key for a given priority.

        For explicit priorities: returns specific allocation and unique pool key
        For default priority: returns default allocation and shared pool key

        Args:
            model: Model name
            priority: Priority level (None for default)
            normalized_weights: Pre-computed normalized weights
            model_info: Model configuration (optional, for fallback conversion)

        Returns:
            tuple: (priority_weight, priority_key)
        """
        # Check if this key has an explicit priority in litellm.priority_reservation
        has_explicit_priority = (
            priority is not None
            and litellm.priority_reservation is not None
            and priority in litellm.priority_reservation
        )

        if has_explicit_priority and priority is not None:
            # Explicit priority: get its specific allocation
            priority_weight = normalized_weights.get(
                priority, self._get_priority_weight(priority, model_info)
            )
            # Use unique key per priority level
            priority_key = f"{model}:{priority}"
        else:
            # No explicit priority: share the default_priority pool with ALL other default keys
            priority_weight = _get_priority_settings().default_priority
            # Use shared key for all default-priority requests
            priority_key = f"{model}:default_pool"

        return priority_weight, priority_key

    async def _check_model_saturation(
        self,
        model: str,
        model_group_info: ModelGroupInfo,
    ) -> float:
        """
        Check current saturation by directly querying v3 limiter's cache keys.

        Reuses v3 limiter's Redis-based tracking (works across multiple instances).
        Reads counters WITHOUT incrementing them.

        Returns:
            float: Saturation ratio (0.0 = empty, 1.0 = at capacity, >1.0 = over)
        """
        try:
            max_saturation = 0.0

            # Query RPM saturation - always read from Redis for multi-node consistency
            if model_group_info.rpm is not None and model_group_info.rpm > 0:
                # Use v3 limiter's key format: {key:value}:rate_limit_type
                counter_key = self.v3_limiter.create_rate_limit_keys(
                    key="model_saturation_check",
                    value=model,
                    rate_limit_type="requests",
                )

                # Query Redis directly for current counter value (skip local cache for consistency)
                counter_value = await self._get_saturation_value_from_cache(
                    counter_key=counter_key
                )

                if counter_value is not None:
                    current_requests = int(counter_value)
                    rpm_saturation = current_requests / model_group_info.rpm
                    max_saturation = max(max_saturation, rpm_saturation)

                    verbose_proxy_logger.debug(
                        f"Model {model} RPM: {current_requests}/{model_group_info.rpm} "
                        f"({rpm_saturation:.1%})"
                    )

            # Query TPM saturation
            if model_group_info.tpm is not None and model_group_info.tpm > 0:
                counter_key = self.v3_limiter.create_rate_limit_keys(
                    key="model_saturation_check",
                    value=model,
                    rate_limit_type="tokens",
                )

                counter_value = await self._get_saturation_value_from_cache(
                    counter_key=counter_key
                )

                if counter_value is not None:
                    current_tokens = float(counter_value)
                    tpm_saturation = current_tokens / model_group_info.tpm
                    max_saturation = max(max_saturation, tpm_saturation)

                    verbose_proxy_logger.debug(
                        f"Model {model} TPM: {current_tokens}/{model_group_info.tpm} "
                        f"({tpm_saturation:.1%})"
                    )

            verbose_proxy_logger.debug(
                f"Model {model} overall saturation: {max_saturation:.1%}"
            )

            return max_saturation

        except Exception as e:
            verbose_proxy_logger.error(
                f"Error checking saturation for {model}: {str(e)}"
            )
            # Fail open: assume not saturated on error
            return 0.0

    def _create_priority_based_descriptors(
        self,
        model: str,
        user_api_key_dict: UserAPIKeyAuth,
        priority: Optional[str],
    ) -> List[RateLimitDescriptor]:
        """
        Create rate limit descriptors with normalized priority weights.

        Uses normalized weights to handle over-allocation scenarios.

        For explicit priorities: each priority gets its own pool (e.g., prod gets 75%)
        For default priority: ALL keys without explicit priority share ONE pool (e.g., all share 25%)
        """
        descriptors: List[RateLimitDescriptor] = []

        if litellm.priority_reservation is None:
            return descriptors

        # Get model group info
        model_group_info: Optional[
            ModelGroupInfo
        ] = self.llm_router.get_model_group_info(model_group=model)
        if model_group_info is None:
            return descriptors

        # Get normalized priority weight and pool key
        normalized_weights = self._normalize_priority_weights(model_group_info)
        priority_weight, priority_key = self._get_priority_allocation(
            model=model,
            priority=priority,
            normalized_weights=normalized_weights,
            model_info=model_group_info,
        )

        rate_limit_config: RateLimitDescriptorRateLimitObject = {}

        # Apply priority weight to model limits
        if model_group_info.tpm is not None:
            reserved_tpm = int(model_group_info.tpm * priority_weight)
            rate_limit_config["tokens_per_unit"] = reserved_tpm

        if model_group_info.rpm is not None:
            reserved_rpm = int(model_group_info.rpm * priority_weight)
            rate_limit_config["requests_per_unit"] = reserved_rpm

        if rate_limit_config:
            rate_limit_config["window_size"] = self.v3_limiter.window_size

            descriptors.append(
                RateLimitDescriptor(
                    key="priority_model",
                    value=priority_key,
                    rate_limit=rate_limit_config,
                )
            )

        return descriptors

    def _create_model_tracking_descriptor(
        self,
        model: str,
        model_group_info: ModelGroupInfo,
        high_limit_multiplier: int = 1,
    ) -> RateLimitDescriptor:
        """
        Create a descriptor for tracking model-wide usage.

        Args:
            model: Model name
            model_group_info: Model configuration with RPM/TPM limits
            high_limit_multiplier: Multiplier for limits (use >1 for tracking-only)

        Returns:
            Rate limit descriptor for model-wide tracking
        """
        return RateLimitDescriptor(
            key="model_saturation_check",
            value=model,
            rate_limit={
                "requests_per_unit": (
                    model_group_info.rpm * high_limit_multiplier
                    if model_group_info.rpm
                    else None
                ),
                "tokens_per_unit": (
                    model_group_info.tpm * high_limit_multiplier
                    if model_group_info.tpm
                    else None
                ),
                "window_size": self.v3_limiter.window_size,
            },
        )

    async def _check_rate_limits(
        self,
        model: str,
        model_group_info: ModelGroupInfo,
        user_api_key_dict: UserAPIKeyAuth,
        priority: Optional[str],
        saturation: float,
        data: dict,
    ) -> None:
        """
        Check rate limits using THREE-PHASE approach to prevent partial increments.

        Phase 1: Read-only check of ALL limits (no increments)
        Phase 2: Decide which limits to enforce based on saturation
        Phase 3: Increment ALL counters atomically (model + priority)

        This prevents the bug where:
        - Model counter increments in stage 1
        - Priority check fails in stage 2
        - Request blocked but model counter already incremented

        Key behaviors:
        - All checks performed first (read-only)
        - Only increment counters if request will be allowed
        - Model capacity: Always enforced at 100%
        - Priority limits: Only enforced when saturated >= threshold
        - Both counters tracked from first request (accurate accounting)

        Args:
            model: Model name
            model_group_info: Model configuration
            user_api_key_dict: User authentication info
            priority: User's priority level
            saturation: Current saturation level
            data: Request data dictionary

        Raises:
            HTTPException: If any limit is exceeded
        """
        import json

        saturation_threshold = _get_priority_settings().saturation_threshold
        should_enforce_priority = saturation >= saturation_threshold

        # Build ALL descriptors upfront
        descriptors_to_check: List[RateLimitDescriptor] = []

        # Model-wide descriptor (always enforce)
        model_wide_descriptor = self._create_model_tracking_descriptor(
            model=model,
            model_group_info=model_group_info,
            high_limit_multiplier=1,
        )
        descriptors_to_check.append(model_wide_descriptor)

        # Priority descriptors (always track, conditionally enforce)
        priority_descriptors = self._create_priority_based_descriptors(
            model=model,
            user_api_key_dict=user_api_key_dict,
            priority=priority,
        )
        if priority_descriptors:
            descriptors_to_check.extend(priority_descriptors)

        # PHASE 1: Read-only check of ALL limits (no increments)
        check_response = await self.v3_limiter.should_rate_limit(
            descriptors=descriptors_to_check,
            parent_otel_span=user_api_key_dict.parent_otel_span,
            read_only=True,  # CRITICAL: Don't increment counters yet
        )

        verbose_proxy_logger.debug(
            f"Read-only check: {json.dumps(check_response, indent=2)}"
        )

        # PHASE 2: Decide which limits to enforce
        if check_response["overall_code"] == "OVER_LIMIT":
            for status in check_response["statuses"]:
                if status["code"] == "OVER_LIMIT":
                    descriptor_key = status["descriptor_key"]

                    # Model-wide limit exceeded (ALWAYS enforce)
                    if descriptor_key == "model_saturation_check":
                        raise HTTPException(
                            status_code=429,
                            detail={
                                "error": f"Model capacity reached for {model}. "
                                f"Priority: {priority}, "
                                f"Rate limit type: {status['rate_limit_type']}, "
                                f"Remaining: {status['limit_remaining']}"
                            },
                            headers={
                                "retry-after": str(self.v3_limiter.window_size),
                                "rate_limit_type": str(status["rate_limit_type"]),
                                "x-litellm-priority": priority or "default",
                            },
                        )

                    # Priority limit exceeded (ONLY enforce when saturated)
                    elif descriptor_key == "priority_model" and should_enforce_priority:
                        verbose_proxy_logger.debug(
                            f"Enforcing priority limits for {model}, saturation: {saturation:.1%}, "
                            f"priority: {priority}"
                        )
                        raise HTTPException(
                            status_code=429,
                            detail={
                                "error": f"Priority-based rate limit exceeded. "
                                f"Priority: {priority}, "
                                f"Rate limit type: {status['rate_limit_type']}, "
                                f"Remaining: {status['limit_remaining']}, "
                                f"Model saturation: {saturation:.1%}"
                            },
                            headers={
                                "retry-after": str(self.v3_limiter.window_size),
                                "rate_limit_type": str(status["rate_limit_type"]),
                                "x-litellm-priority": priority or "default",
                                "x-litellm-saturation": f"{saturation:.2%}",
                            },
                        )

        # PHASE 3: Increment counters separately to avoid early-exit issues
        # Model counter must ALWAYS increment, but priority counter might be over limit
        # If we increment them together, v3_limiter's in-memory check will exit early
        # and skip incrementing the model counter

        # Step 3a: Increment model-wide counter (always)
        model_increment_response = await self.v3_limiter.should_rate_limit(
            descriptors=[model_wide_descriptor],
            parent_otel_span=user_api_key_dict.parent_otel_span,
            read_only=False,
        )

        # Step 3b: Increment priority counter (may be over limit, but we still track it)
        if priority_descriptors:
            priority_increment_response = await self.v3_limiter.should_rate_limit(
                descriptors=priority_descriptors,
                parent_otel_span=user_api_key_dict.parent_otel_span,
                read_only=False,
            )

            # Combine responses for post-call hook
            combined_response = {
                "overall_code": model_increment_response["overall_code"],
                "statuses": model_increment_response["statuses"]
                + priority_increment_response["statuses"],
            }
            data["litellm_proxy_rate_limit_response"] = combined_response
        else:
            data["litellm_proxy_rate_limit_response"] = model_increment_response

    async def async_pre_call_hook(
        self,
        user_api_key_dict: UserAPIKeyAuth,
        cache: DualCache,
        data: dict,
        call_type: CallTypesLiteral,
    ) -> Optional[Union[Exception, str, dict]]:
        """
        Saturation-aware pre-call hook for priority-based rate limiting.

        Flow:
        1. Check current saturation level
        2. THREE-PHASE rate limit check:
           - PHASE 1: Read-only check of ALL limits (no increments)
           - PHASE 2: Decide which limits to enforce based on saturation
           - PHASE 3: Increment ALL counters atomically if request allowed

        This three-phase approach ensures:
        - Model capacity is NEVER exceeded (always enforced at 100%)
        - Priority usage tracked from first request (accurate metrics)
        - Counters only increment when request will be allowed (prevents phantom usage)
        - When under-saturated: priorities can borrow unused capacity (generous)
        - When saturated: fair allocation based on normalized priority weights (strict)

        Example with 100 RPM model, 60% priority allocation, 80% threshold:
        - Saturation < 80%: Priority can use up to 100 RPM (model limit enforced only)
        - Saturation >= 80%: Priority limited to 60 RPM (both limits enforced)

        Prevents bugs where:
        - Model counter increments but priority check fails → model over-capacity
        - Priority counter increments but not enforced → inaccurate metrics

        Args:
            user_api_key_dict: User authentication and metadata
            cache: Dual cache instance
            data: Request data containing model name
            call_type: Type of API call being made

        Returns:
            None if request is allowed, otherwise raises HTTPException
        """
        if "model" not in data:
            return None

        model = data["model"]
        priority = self._get_priority_from_user_api_key_dict(
            user_api_key_dict=user_api_key_dict
        )

        # Get model configuration
        model_group_info: Optional[
            ModelGroupInfo
        ] = self.llm_router.get_model_group_info(model_group=model)
        if model_group_info is None:
            verbose_proxy_logger.debug(
                f"No model group info for {model}, allowing request"
            )
            return None

        try:
            # STEP 1: Check current saturation level
            saturation = await self._check_model_saturation(model, model_group_info)

            saturation_threshold = _get_priority_settings().saturation_threshold

            verbose_proxy_logger.debug(
                f"[Dynamic Rate Limiter] Model={model}, Saturation={saturation:.1%}, "
                f"Threshold={saturation_threshold:.1%}, Priority={priority}"
            )

            # STEP 2: Check rate limits in THREE phases
            # Phase 1: Read-only check of ALL limits (no increments)
            # Phase 2: Decide which limits to enforce (based on saturation)
            # Phase 3: Increment ALL counters only if request will be allowed
            # This prevents partial increments and ensures accurate tracking
            await self._check_rate_limits(
                model=model,
                model_group_info=model_group_info,
                user_api_key_dict=user_api_key_dict,
                priority=priority,
                saturation=saturation,
                data=data,
            )

        except HTTPException:
            raise
        except Exception as e:
            verbose_proxy_logger.error(
                f"Error in dynamic rate limiter: {str(e)}, allowing request"
            )
            # Fail open on unexpected errors
            return None

        return None

    async def async_post_call_success_hook(
        self, data: dict, user_api_key_dict: UserAPIKeyAuth, response
    ):
        """
        Post-call hook to add rate limit headers to response.
        Leverages v3 limiter's post-call hook functionality.
        """
        try:
            # Call v3 limiter's post-call hook to add standard rate limit headers
            await self.v3_limiter.async_post_call_success_hook(
                data=data, user_api_key_dict=user_api_key_dict, response=response
            )

            # Add additional priority-specific headers
            if isinstance(response, ModelResponse):
                priority = self._get_priority_from_user_api_key_dict(
                    user_api_key_dict=user_api_key_dict
                )

                # Get existing additional headers
                additional_headers = (
                    getattr(response, "_hidden_params", {}).get(
                        "additional_headers", {}
                    )
                    or {}
                )

                # Add priority information
                additional_headers["x-litellm-priority"] = priority or "default"
                additional_headers["x-litellm-rate-limiter-version"] = "v3"

                # Update response
                if not hasattr(response, "_hidden_params"):
                    response._hidden_params = {}
                response._hidden_params["additional_headers"] = additional_headers

            return response

        except Exception as e:
            verbose_proxy_logger.exception(
                f"Error in dynamic rate limiter v3 post-call hook: {str(e)}"
            )
            return response

    async def async_log_success_event(self, kwargs, response_obj, start_time, end_time):
        """
        Update token usage for priority-based rate limiting after successful API calls.

        Increments token counters for:
        - model_saturation_check: Model-wide token tracking
        - priority_model: Priority-specific token tracking
        """
        from litellm.litellm_core_utils.core_helpers import (
            _get_parent_otel_span_from_kwargs,
        )
        from litellm.proxy.common_utils.callback_utils import (
            get_model_group_from_litellm_kwargs,
        )
        from litellm.types.caching import RedisPipelineIncrementOperation
        from litellm.types.utils import Usage

        try:
            verbose_proxy_logger.debug(
                "INSIDE dynamic rate limiter ASYNC SUCCESS LOGGING"
            )

            litellm_parent_otel_span = _get_parent_otel_span_from_kwargs(kwargs)

            # Get metadata from standard_logging_object
            standard_logging_object = kwargs.get("standard_logging_object") or {}
            standard_logging_metadata = standard_logging_object.get("metadata") or {}

            # Get model and priority
            model_group = get_model_group_from_litellm_kwargs(kwargs)
            if not model_group:
                return

            # Get priority from user_api_key_auth_metadata in standard_logging_metadata
            # This is where user_api_key_dict.metadata is stored during pre-call
            user_api_key_auth_metadata = (
                standard_logging_metadata.get("user_api_key_auth_metadata") or {}
            )
            key_priority: Optional[str] = user_api_key_auth_metadata.get("priority")

            # Get total tokens from response
            total_tokens = 0
            rate_limit_type = self.v3_limiter.get_rate_limit_type()

            if isinstance(response_obj, ModelResponse):
                _usage = getattr(response_obj, "usage", None)
                if _usage and isinstance(_usage, Usage):
                    if rate_limit_type == "output":
                        total_tokens = _usage.completion_tokens
                    elif rate_limit_type == "input":
                        total_tokens = _usage.prompt_tokens
                    elif rate_limit_type == "total":
                        total_tokens = _usage.total_tokens

            if total_tokens == 0:
                return

            # Create pipeline operations for token increments
            pipeline_operations: List[RedisPipelineIncrementOperation] = []

            # Model-wide token tracking (model_saturation_check)
            model_token_key = self.v3_limiter.create_rate_limit_keys(
                key="model_saturation_check",
                value=model_group,
                rate_limit_type="tokens",
            )
            pipeline_operations.append(
                RedisPipelineIncrementOperation(
                    key=model_token_key,
                    increment_value=total_tokens,
                    ttl=self.v3_limiter.window_size,
                )
            )

            # Priority-specific token tracking (priority_model)
            # Determine priority key (same logic as _get_priority_allocation)
            has_explicit_priority = (
                key_priority is not None
                and litellm.priority_reservation is not None
                and key_priority in litellm.priority_reservation
            )

            if has_explicit_priority and key_priority is not None:
                priority_key = f"{model_group}:{key_priority}"
            else:
                priority_key = f"{model_group}:default_pool"

            priority_token_key = self.v3_limiter.create_rate_limit_keys(
                key="priority_model",
                value=priority_key,
                rate_limit_type="tokens",
            )
            pipeline_operations.append(
                RedisPipelineIncrementOperation(
                    key=priority_token_key,
                    increment_value=total_tokens,
                    ttl=self.v3_limiter.window_size,
                )
            )

            # Execute token increments with TTL preservation
            if pipeline_operations:
                await self.v3_limiter.async_increment_tokens_with_ttl_preservation(
                    pipeline_operations=pipeline_operations,
                    parent_otel_span=litellm_parent_otel_span,
                )

                # Only log 'priority' if it's known safe; otherwise, redact.
                SAFE_PRIORITIES = {"low", "medium", "high", "default"}
                logged_priority = (
                    key_priority if key_priority in SAFE_PRIORITIES else "REDACTED"
                )
                verbose_proxy_logger.debug(
                    f"[Dynamic Rate Limiter] Incremented tokens by {total_tokens} for "
                    f"model={model_group}, priority={logged_priority}"
                )

        except Exception as e:
            verbose_proxy_logger.exception(
                f"Error in dynamic rate limiter success event: {str(e)}"
            )



import litellm
from litellm.llms.openai.chat.gpt_transformation import OpenAIGPTConfig


class VertexAIGPTOSSTransformation(OpenAIGPTConfig):
    """
    Transformation for GPT-OSS model from VertexAI

    https://console.cloud.google.com/vertex-ai/publishers/openai/model-garden/gpt-oss-120b-maas?hl=id
    """
    def __init__(self):
        super().__init__()
    
    def get_supported_openai_params(self, model: str) -> list:
        base_gpt_series_params = super().get_supported_openai_params(model=model)
        gpt_oss_only_params = ["reasoning_effort"]
        base_gpt_series_params.extend(gpt_oss_only_params)

        #########################################################
        # VertexAI - GPT-OSS does not support tool calls
        #########################################################
        if litellm.supports_function_calling(model=model) is False:
            TOOL_CALLING_PARAMS_TO_REMOVE = ["tool", "tool_choice", "function_call", "functions"]
            base_gpt_series_params = [param for param in base_gpt_series_params if param not in TOOL_CALLING_PARAMS_TO_REMOVE]

        return base_gpt_series_params




"""Google AI Studio Interactions API implementation."""

from litellm.llms.gemini.interactions.transformation import (
    GoogleAIStudioInteractionsConfig,
)

__all__ = ["GoogleAIStudioInteractionsConfig"]



"""
Pass-Through Endpoint Guardrail Translation

This module exists here (under litellm/llms/) so it can be auto-discovered by
load_guardrail_translation_mappings() which scans for guardrail_translation
directories under litellm/llms/.

The main passthrough endpoint implementation is in:
    litellm/proxy/pass_through_endpoints/

See guardrail_translation/README.md for more details.
"""



"""
Passthrough Guardrails Helper Module

Handles guardrail execution for passthrough endpoints with:
- Opt-in model (guardrails only run when explicitly configured)
- Field-level targeting using JSONPath expressions
- Automatic inheritance from org/team/key levels when enabled
"""

from typing import Any, Dict, List, Optional, Union

from litellm._logging import verbose_proxy_logger
from litellm.proxy._types import (
    PassThroughGuardrailsConfig,
    PassThroughGuardrailSettings,
    UserAPIKeyAuth,
)
from litellm.proxy.pass_through_endpoints.jsonpath_extractor import JsonPathExtractor

# Type for raw guardrails config input (before normalization)
# Can be a list of names or a dict with settings
PassThroughGuardrailsConfigInput = Union[
    List[str],  # Simple list: ["guard-1", "guard-2"]
    PassThroughGuardrailsConfig,  # Dict: {"guard-1": {"request_fields": [...]}}
]


class PassthroughGuardrailHandler:
    """
    Handles guardrail execution for passthrough endpoints.
    
    Passthrough endpoints use an opt-in model for guardrails:
    - Guardrails only run when explicitly configured on the endpoint
    - Supports field-level targeting using JSONPath expressions
    - Automatically inherits org/team/key level guardrails when enabled
    
    Guardrails can be specified as:
    - List format (simple): ["guardrail-1", "guardrail-2"]
    - Dict format (with settings): {"guardrail-1": {"request_fields": ["query"]}}
    """

    @staticmethod
    def normalize_config(
        guardrails_config: Optional[PassThroughGuardrailsConfigInput],
    ) -> Optional[PassThroughGuardrailsConfig]:
        """
        Normalize guardrails config to dict format.
        
        Accepts:
        - List of guardrail names: ["g1", "g2"] -> {"g1": None, "g2": None}
        - Dict with settings: {"g1": {"request_fields": [...]}}
        - None: returns None
        """
        if guardrails_config is None:
            return None
        
        # Already a dict - return as-is
        if isinstance(guardrails_config, dict):
            return guardrails_config
        
        # List of guardrail names - convert to dict
        if isinstance(guardrails_config, list):
            return {name: None for name in guardrails_config}
        
        verbose_proxy_logger.debug(
            "Passthrough guardrails config is not a dict or list, got: %s",
            type(guardrails_config),
        )
        return None

    @staticmethod
    def is_enabled(
        guardrails_config: Optional[PassThroughGuardrailsConfigInput],
    ) -> bool:
        """
        Check if guardrails are enabled for a passthrough endpoint.
        
        Passthrough endpoints are opt-in only - guardrails only run when 
        the guardrails config is set with at least one guardrail.
        """
        normalized = PassthroughGuardrailHandler.normalize_config(guardrails_config)
        if normalized is None:
            return False
        return len(normalized) > 0

    @staticmethod
    def get_guardrail_names(
        guardrails_config: Optional[PassThroughGuardrailsConfigInput],
    ) -> List[str]:
        """Get the list of guardrail names configured for a passthrough endpoint."""
        normalized = PassthroughGuardrailHandler.normalize_config(guardrails_config)
        if normalized is None:
            return []
        return list(normalized.keys())

    @staticmethod
    def get_settings(
        guardrails_config: Optional[PassThroughGuardrailsConfigInput],
        guardrail_name: str,
    ) -> Optional[PassThroughGuardrailSettings]:
        """Get settings for a specific guardrail from the passthrough config."""
        normalized = PassthroughGuardrailHandler.normalize_config(guardrails_config)
        if normalized is None:
            return None
        
        settings = normalized.get(guardrail_name)
        if settings is None:
            return None
        
        if isinstance(settings, dict):
            return PassThroughGuardrailSettings(**settings)
        
        return settings

    @staticmethod
    def prepare_input(
        request_data: dict,
        guardrail_settings: Optional[PassThroughGuardrailSettings],
    ) -> str:
        """
        Prepare input text for guardrail execution based on field targeting settings.
        
        If request_fields is specified, extracts only those fields.
        Otherwise, uses the entire request payload as text.
        """
        if guardrail_settings is None or guardrail_settings.request_fields is None:
            from litellm.litellm_core_utils.safe_json_dumps import safe_dumps
            return safe_dumps(request_data)
        
        return JsonPathExtractor.extract_fields(
            data=request_data,
            jsonpath_expressions=guardrail_settings.request_fields,
        )

    @staticmethod
    def prepare_output(
        response_data: dict,
        guardrail_settings: Optional[PassThroughGuardrailSettings],
    ) -> str:
        """
        Prepare output text for guardrail execution based on field targeting settings.
        
        If response_fields is specified, extracts only those fields.
        Otherwise, uses the entire response payload as text.
        """
        if guardrail_settings is None or guardrail_settings.response_fields is None:
            from litellm.litellm_core_utils.safe_json_dumps import safe_dumps
            return safe_dumps(response_data)
        
        return JsonPathExtractor.extract_fields(
            data=response_data,
            jsonpath_expressions=guardrail_settings.response_fields,
        )

    @staticmethod
    async def execute(
        request_data: dict,
        user_api_key_dict: UserAPIKeyAuth,
        guardrails_config: Optional[PassThroughGuardrailsConfig],
        event_type: str = "pre_call",
    ) -> dict:
        """
        Execute guardrails for a passthrough endpoint.
        
        This is the main entry point for passthrough guardrail execution.
        
        Args:
            request_data: The request payload
            user_api_key_dict: User API key authentication info
            guardrails_config: Passthrough-specific guardrails configuration
            event_type: "pre_call" for request, "post_call" for response
        
        Returns:
            The potentially modified request_data
        
        Raises:
            HTTPException if a guardrail blocks the request
        """
        if not PassthroughGuardrailHandler.is_enabled(guardrails_config):
            verbose_proxy_logger.debug(
                "Passthrough guardrails not enabled, skipping guardrail execution"
            )
            return request_data
        
        guardrail_names = PassthroughGuardrailHandler.get_guardrail_names(
            guardrails_config
        )
        verbose_proxy_logger.debug(
            "Executing passthrough guardrails: %s", guardrail_names
        )
        
        # Add to request metadata so guardrails know which to run
        from litellm.proxy.pass_through_endpoints.passthrough_context import (
            set_passthrough_guardrails_config,
        )

        if "metadata" not in request_data:
            request_data["metadata"] = {}
        
        # Set guardrails in metadata using dict format for compatibility
        request_data["metadata"]["guardrails"] = {
            name: True for name in guardrail_names
        }
        
        # Store passthrough guardrails config in request-scoped context
        set_passthrough_guardrails_config(guardrails_config)
        
        return request_data

    @staticmethod
    def collect_guardrails(
        user_api_key_dict: UserAPIKeyAuth,
        passthrough_guardrails_config: Optional[PassThroughGuardrailsConfigInput],
    ) -> Optional[Dict[str, bool]]:
        """
        Collect guardrails for a passthrough endpoint.

        Passthrough endpoints are opt-in only for guardrails. Guardrails only run when
        the guardrails config is set with at least one guardrail.

        Accepts both list and dict formats:
        - List: ["guardrail-1", "guardrail-2"]
        - Dict: {"guardrail-1": {"request_fields": [...]}}

        When enabled, this function collects:
        - Passthrough-specific guardrails from the config
        - Org/team/key level guardrails (automatic inheritance when passthrough is enabled)

        Args:
            user_api_key_dict: User API key authentication info
            passthrough_guardrails_config: List or Dict of guardrail names/settings

        Returns:
            Dict of guardrail names to run (format: {guardrail_name: True}), or None
        """
        from litellm.proxy.litellm_pre_call_utils import (
            _add_guardrails_from_key_or_team_metadata,
        )

        # Normalize config to dict format (handles both list and dict)
        normalized_config = PassthroughGuardrailHandler.normalize_config(
            passthrough_guardrails_config
        )

        if normalized_config is None:
            verbose_proxy_logger.debug(
                "Passthrough guardrails not configured, skipping guardrail collection"
            )
            return None

        if len(normalized_config) == 0:
            verbose_proxy_logger.debug(
                "Passthrough guardrails config is empty, skipping"
            )
            return None

        # Passthrough is enabled - collect guardrails
        guardrails_to_run: Dict[str, bool] = {}

        # Add passthrough-specific guardrails
        for guardrail_name in normalized_config.keys():
            guardrails_to_run[guardrail_name] = True
            verbose_proxy_logger.debug(
                "Added passthrough-specific guardrail: %s", guardrail_name
            )

        # Add org/team/key level guardrails using shared helper
        temp_data: Dict[str, Any] = {"metadata": {}}
        _add_guardrails_from_key_or_team_metadata(
            key_metadata=user_api_key_dict.metadata,
            team_metadata=user_api_key_dict.team_metadata,
            data=temp_data,
            metadata_variable_name="metadata",
        )

        # Merge inherited guardrails into guardrails_to_run
        inherited_guardrails = temp_data["metadata"].get("guardrails", [])
        for guardrail_name in inherited_guardrails:
            if guardrail_name not in guardrails_to_run:
                guardrails_to_run[guardrail_name] = True
                verbose_proxy_logger.debug(
                    "Added inherited guardrail (key/team level): %s", guardrail_name
                )

        verbose_proxy_logger.debug(
            "Collected guardrails for passthrough endpoint: %s",
            list(guardrails_to_run.keys()),
        )

        return guardrails_to_run if guardrails_to_run else None

    @staticmethod
    def get_field_targeted_text(
        data: dict,
        guardrail_name: str,
        is_request: bool = True,
    ) -> Optional[str]:
        """
        Get the text to check for a guardrail, respecting field targeting settings.
        
        Called by guardrail hooks to get the appropriate text based on
        passthrough field targeting configuration.
        
        Args:
            data: The request/response data dict
            guardrail_name: Name of the guardrail being executed
            is_request: True for request (pre_call), False for response (post_call)
        
        Returns:
            The text to check, or None to use default behavior
        """
        from litellm.proxy.pass_through_endpoints.passthrough_context import (
            get_passthrough_guardrails_config,
        )

        passthrough_config = get_passthrough_guardrails_config()
        if passthrough_config is None:
            return None
        
        settings = PassthroughGuardrailHandler.get_settings(
            passthrough_config, guardrail_name
        )
        if settings is None:
            return None
        
        if is_request:
            if settings.request_fields:
                return JsonPathExtractor.extract_fields(data, settings.request_fields)
        else:
            if settings.response_fields:
                return JsonPathExtractor.extract_fields(data, settings.response_fields)
        
        return None



"""
Registry data for lazy imports.

This module contains all the name tuples and import maps used by the lazy import system.
Separated from the handler functions for better organization.
"""

# Cost calculator names that support lazy loading via _lazy_import_cost_calculator
COST_CALCULATOR_NAMES = (
    "completion_cost",
    "cost_per_token",
    "response_cost_calculator",
)

# Litellm logging names that support lazy loading via _lazy_import_litellm_logging
LITELLM_LOGGING_NAMES = (
    "Logging",
    "modify_integration",
)

# Utils names that support lazy loading via _lazy_import_utils
UTILS_NAMES = (
    "exception_type",
    "get_optional_params",
    "get_response_string",
    "token_counter",
    "create_pretrained_tokenizer",
    "create_tokenizer",
    "supports_function_calling",
    "supports_web_search",
    "supports_url_context",
    "supports_response_schema",
    "supports_parallel_function_calling",
    "supports_vision",
    "supports_audio_input",
    "supports_audio_output",
    "supports_system_messages",
    "supports_reasoning",
    "get_litellm_params",
    "acreate",
    "get_max_tokens",
    "get_model_info",
    "register_prompt_template",
    "validate_environment",
    "check_valid_key",
    "register_model",
    "encode",
    "decode",
    "_calculate_retry_after",
    "_should_retry",
    "get_supported_openai_params",
    "get_api_base",
    "get_first_chars_messages",
    "ModelResponse",
    "ModelResponseStream",
    "EmbeddingResponse",
    "ImageResponse",
    "TranscriptionResponse",
    "TextCompletionResponse",
    "get_provider_fields",
    "ModelResponseListIterator",
    "get_valid_models",
    "timeout",
    "get_llm_provider",
    "remove_index_from_tool_calls",
)

# Token counter names that support lazy loading via _lazy_import_token_counter
TOKEN_COUNTER_NAMES = ("get_modified_max_tokens",)

# LLM client cache names that support lazy loading via _lazy_import_llm_client_cache
LLM_CLIENT_CACHE_NAMES = (
    "LLMClientCache",
    "in_memory_llm_clients_cache",
)

# Bedrock type names that support lazy loading via _lazy_import_bedrock_types
BEDROCK_TYPES_NAMES = ("COHERE_EMBEDDING_INPUT_TYPES",)

# Common types from litellm.types.utils that support lazy loading via
# _lazy_import_types_utils
TYPES_UTILS_NAMES = (
    "ImageObject",
    "BudgetConfig",
    "all_litellm_params",
    "_litellm_completion_params",
    "CredentialItem",
    "PriorityReservationDict",
    "StandardKeyGenerationConfig",
    "SearchProviders",
    "GenericStreamingChunk",
)

# Caching / cache classes that support lazy loading via _lazy_import_caching
CACHING_NAMES = (
    "Cache",
    "DualCache",
    "RedisCache",
    "InMemoryCache",
)

# HTTP handler names that support lazy loading via _lazy_import_http_handlers
HTTP_HANDLER_NAMES = (
    "module_level_aclient",
    "module_level_client",
)

# Dotprompt integration names that support lazy loading via _lazy_import_dotprompt
DOTPROMPT_NAMES = (
    "global_prompt_manager",
    "global_prompt_directory",
    "set_global_prompt_directory",
)

# LLM config classes that support lazy loading via _lazy_import_llm_configs
LLM_CONFIG_NAMES = (
    "AmazonConverseConfig",
    "OpenAILikeChatConfig",
    "GaladrielChatConfig",
    "GithubChatConfig",
    "AzureAnthropicConfig",
    "BytezChatConfig",
    "CompactifAIChatConfig",
    "EmpowerChatConfig",
    "MinimaxChatConfig",
    "AiohttpOpenAIChatConfig",
    "HuggingFaceChatConfig",
    "HuggingFaceEmbeddingConfig",
    "OobaboogaConfig",
    "MaritalkConfig",
    "OpenrouterConfig",
    "DataRobotConfig",
    "AnthropicConfig",
    "AnthropicTextConfig",
    "GroqSTTConfig",
    "TritonConfig",
    "TritonGenerateConfig",
    "TritonInferConfig",
    "TritonEmbeddingConfig",
    "HuggingFaceRerankConfig",
    "DatabricksConfig",
    "DatabricksEmbeddingConfig",
    "PredibaseConfig",
    "ReplicateConfig",
    "SnowflakeConfig",
    "CohereRerankConfig",
    "CohereRerankV2Config",
    "AzureAIRerankConfig",
    "InfinityRerankConfig",
    "JinaAIRerankConfig",
    "DeepinfraRerankConfig",
    "HostedVLLMRerankConfig",
    "NvidiaNimRerankConfig",
    "NvidiaNimRankingConfig",
    "VertexAIRerankConfig",
    "FireworksAIRerankConfig",
    "VoyageRerankConfig",
    "ClarifaiConfig",
    "AI21ChatConfig",
    "LlamaAPIConfig",
    "TogetherAITextCompletionConfig",
    "CloudflareChatConfig",
    "NovitaConfig",
    "PetalsConfig",
    "OllamaChatConfig",
    "OllamaConfig",
    "SagemakerConfig",
    "SagemakerChatConfig",
    "CohereChatConfig",
    "AnthropicMessagesConfig",
    "AmazonAnthropicClaudeMessagesConfig",
    "TogetherAIConfig",
    "NLPCloudConfig",
    "VertexGeminiConfig",
    "GoogleAIStudioGeminiConfig",
    "VertexAIAnthropicConfig",
    "VertexAILlama3Config",
    "VertexAIAi21Config",
    "AmazonCohereChatConfig",
    "AmazonBedrockGlobalConfig",
    "AmazonAI21Config",
    "AmazonInvokeNovaConfig",
    "AmazonQwen2Config",
    "AmazonQwen3Config",
    # Aliases for backwards compatibility
    "VertexAIConfig",  # Alias for VertexGeminiConfig
    "GeminiConfig",  # Alias for GoogleAIStudioGeminiConfig
    "AmazonAnthropicConfig",
    "AmazonAnthropicClaudeConfig",
    "AmazonCohereConfig",
    "AmazonLlamaConfig",
    "AmazonDeepSeekR1Config",
    "AmazonMistralConfig",
    "AmazonMoonshotConfig",
    "AmazonTitanConfig",
    "AmazonTwelveLabsPegasusConfig",
    "AmazonInvokeConfig",
    "AmazonBedrockOpenAIConfig",
    "AmazonStabilityConfig",
    "AmazonStability3Config",
    "AmazonNovaCanvasConfig",
    "AmazonTitanG1Config",
    "AmazonTitanMultimodalEmbeddingG1Config",
    "CohereV2ChatConfig",
    "BedrockCohereEmbeddingConfig",
    "TwelveLabsMarengoEmbeddingConfig",
    "AmazonNovaEmbeddingConfig",
    "OpenAIConfig",
    "MistralEmbeddingConfig",
    "OpenAIImageVariationConfig",
    "DeepInfraConfig",
    "DeepgramAudioTranscriptionConfig",
    "TopazImageVariationConfig",
    "OpenAITextCompletionConfig",
    "GroqChatConfig",
    "A2AConfig",
    "GenAIHubOrchestrationConfig",
    "VoyageEmbeddingConfig",
    "VoyageContextualEmbeddingConfig",
    "InfinityEmbeddingConfig",
    "AzureAIStudioConfig",
    "MistralConfig",
    "OpenAIResponsesAPIConfig",
    "AzureOpenAIResponsesAPIConfig",
    "AzureOpenAIOSeriesResponsesAPIConfig",
    "XAIResponsesAPIConfig",
    "LiteLLMProxyResponsesAPIConfig",
    "VolcEngineResponsesAPIConfig",
    "PerplexityResponsesConfig",
    "GoogleAIStudioInteractionsConfig",
    "OpenAIOSeriesConfig",
    "AnthropicSkillsConfig",
    "BaseSkillsAPIConfig",
    "GradientAIConfig",
    # Alias for backwards compatibility
    "OpenAIO1Config",  # Alias for OpenAIOSeriesConfig
    "OpenAIGPTConfig",
    "OpenAIGPT5Config",
    "OpenAIWhisperAudioTranscriptionConfig",
    "OpenAIGPTAudioTranscriptionConfig",
    "OpenAIGPTAudioConfig",
    "NvidiaNimConfig",
    "NvidiaNimEmbeddingConfig",
    "FeatherlessAIConfig",
    "CerebrasConfig",
    "BasetenConfig",
    "SambanovaConfig",
    "SambaNovaEmbeddingConfig",
    "FireworksAIConfig",
    "FireworksAITextCompletionConfig",
    "FireworksAIAudioTranscriptionConfig",
    "FireworksAIEmbeddingConfig",
    "FriendliaiChatConfig",
    "JinaAIEmbeddingConfig",
    "XAIChatConfig",
    "ZAIChatConfig",
    "AIMLChatConfig",
    "VolcEngineChatConfig",
    "CodestralTextCompletionConfig",
    "AzureOpenAIAssistantsAPIConfig",
    "HerokuChatConfig",
    "CometAPIConfig",
    "AzureOpenAIConfig",
    "AzureOpenAIGPT5Config",
    "AzureOpenAITextConfig",
    "HostedVLLMChatConfig",
    "HostedVLLMEmbeddingConfig",
    # Alias for backwards compatibility
    "VolcEngineConfig",  # Alias for VolcEngineChatConfig
    "LlamafileChatConfig",
    "LiteLLMProxyChatConfig",
    "VLLMConfig",
    "DeepSeekChatConfig",
    "LMStudioChatConfig",
    "LmStudioEmbeddingConfig",
    "NscaleConfig",
    "PerplexityChatConfig",
    "AzureOpenAIO1Config",
    "IBMWatsonXAIConfig",
    "IBMWatsonXChatConfig",
    "IBMWatsonXEmbeddingConfig",
    "GenAIHubEmbeddingConfig",
    "IBMWatsonXAudioTranscriptionConfig",
    "GithubCopilotConfig",
    "GithubCopilotResponsesAPIConfig",
    "ChatGPTConfig",
    "ChatGPTResponsesAPIConfig",
    "ManusResponsesAPIConfig",
    "GithubCopilotEmbeddingConfig",
    "NebiusConfig",
    "WandbConfig",
    "GigaChatConfig",
    "GigaChatEmbeddingConfig",
    "DashScopeChatConfig",
    "MoonshotChatConfig",
    "DockerModelRunnerChatConfig",
    "V0ChatConfig",
    "OCIChatConfig",
    "MorphChatConfig",
    "RAGFlowConfig",
    "LambdaAIChatConfig",
    "HyperbolicChatConfig",
    "VercelAIGatewayConfig",
    "OVHCloudChatConfig",
    "OVHCloudEmbeddingConfig",
    "CometAPIEmbeddingConfig",
    "LemonadeChatConfig",
    "SnowflakeEmbeddingConfig",
    "AmazonNovaChatConfig",
)

# Types that support lazy loading via _lazy_import_types
TYPES_NAMES = (
    "GuardrailItem",
    "DefaultTeamSSOParams",
    "LiteLLM_UpperboundKeyGenerateParams",
    "KeyManagementSystem",
    "PriorityReservationSettings",
    "CustomLogger",
    "LoggingCallbackManager",
    "DatadogLLMObsInitParams",
    # Note: LlmProviders is NOT lazy-loaded because it's imported during import time
    # in multiple places including openai.py (via main import)
    # Note: KeyManagementSettings is NOT lazy-loaded because _key_management_settings
    # is accessed during import time in secret_managers/main.py
)

# LLM provider logic names that support lazy loading via _lazy_import_llm_provider_logic
LLM_PROVIDER_LOGIC_NAMES = (
    "get_llm_provider",
    "remove_index_from_tool_calls",
)

# Utils module names that support lazy loading via _lazy_import_utils_module
# These are attributes accessed from litellm.utils module
UTILS_MODULE_NAMES = (
    "encoding",
    "BaseVectorStore",
    "CredentialAccessor",
    "exception_type",
    "get_error_message",
    "_get_response_headers",
    "get_llm_provider",
    "_is_non_openai_azure_model",
    "get_supported_openai_params",
    "LiteLLMResponseObjectHandler",
    "_handle_invalid_parallel_tool_calls",
    "convert_to_model_response_object",
    "convert_to_streaming_response",
    "convert_to_streaming_response_async",
    "get_api_base",
    "ResponseMetadata",
    "_parse_content_for_reasoning",
    "LiteLLMLoggingObject",
    "redact_message_input_output_from_logging",
    "CustomStreamWrapper",
    "BaseGoogleGenAIGenerateContentConfig",
    "BaseOCRConfig",
    "BaseSearchConfig",
    "BaseTextToSpeechConfig",
    "BedrockModelInfo",
    "CohereModelInfo",
    "MistralOCRConfig",
    "Rules",
    "AsyncHTTPHandler",
    "HTTPHandler",
    "get_num_retries_from_retry_policy",
    "reset_retry_policy",
    "get_secret",
    "get_coroutine_checker",
    "get_litellm_logging_class",
    "get_set_callbacks",
    "get_litellm_metadata_from_kwargs",
    "map_finish_reason",
    "process_response_headers",
    "delete_nested_value",
    "is_nested_path",
    "_get_base_model_from_litellm_call_metadata",
    "get_litellm_params",
    "_ensure_extra_body_is_safe",
    "get_formatted_prompt",
    "get_response_headers",
    "update_response_metadata",
    "executor",
    "BaseAnthropicMessagesConfig",
    "BaseAudioTranscriptionConfig",
    "BaseBatchesConfig",
    "BaseContainerConfig",
    "BaseEmbeddingConfig",
    "BaseImageEditConfig",
    "BaseImageGenerationConfig",
    "BaseImageVariationConfig",
    "BasePassthroughConfig",
    "BaseRealtimeConfig",
    "BaseRerankConfig",
    "BaseVectorStoreConfig",
    "BaseVectorStoreFilesConfig",
    "BaseVideoConfig",
    "ANTHROPIC_API_ONLY_HEADERS",
    "AnthropicThinkingParam",
    "RerankResponse",
    "ChatCompletionDeltaToolCallChunk",
    "ChatCompletionToolCallChunk",
    "ChatCompletionToolCallFunctionChunk",
    "LiteLLM_Params",
)

# Import maps for registry pattern - reduces repetition
_UTILS_IMPORT_MAP = {
    "exception_type": (".utils", "exception_type"),
    "get_optional_params": (".utils", "get_optional_params"),
    "get_response_string": (".utils", "get_response_string"),
    "token_counter": (".utils", "token_counter"),
    "create_pretrained_tokenizer": (".utils", "create_pretrained_tokenizer"),
    "create_tokenizer": (".utils", "create_tokenizer"),
    "supports_function_calling": (".utils", "supports_function_calling"),
    "supports_web_search": (".utils", "supports_web_search"),
    "supports_url_context": (".utils", "supports_url_context"),
    "supports_response_schema": (".utils", "supports_response_schema"),
    "supports_parallel_function_calling": (
        ".utils",
        "supports_parallel_function_calling",
    ),
    "supports_vision": (".utils", "supports_vision"),
    "supports_audio_input": (".utils", "supports_audio_input"),
    "supports_audio_output": (".utils", "supports_audio_output"),
    "supports_system_messages": (".utils", "supports_system_messages"),
    "supports_reasoning": (".utils", "supports_reasoning"),
    "get_litellm_params": (".utils", "get_litellm_params"),
    "acreate": (".utils", "acreate"),
    "get_max_tokens": (".utils", "get_max_tokens"),
    "get_model_info": (".utils", "get_model_info"),
    "register_prompt_template": (".utils", "register_prompt_template"),
    "validate_environment": (".utils", "validate_environment"),
    "check_valid_key": (".utils", "check_valid_key"),
    "register_model": (".utils", "register_model"),
    "encode": (".utils", "encode"),
    "decode": (".utils", "decode"),
    "_calculate_retry_after": (".utils", "_calculate_retry_after"),
    "_should_retry": (".utils", "_should_retry"),
    "get_supported_openai_params": (".utils", "get_supported_openai_params"),
    "get_api_base": (".utils", "get_api_base"),
    "get_first_chars_messages": (".utils", "get_first_chars_messages"),
    "ModelResponse": (".utils", "ModelResponse"),
    "ModelResponseStream": (".utils", "ModelResponseStream"),
    "EmbeddingResponse": (".utils", "EmbeddingResponse"),
    "ImageResponse": (".utils", "ImageResponse"),
    "TranscriptionResponse": (".utils", "TranscriptionResponse"),
    "TextCompletionResponse": (".utils", "TextCompletionResponse"),
    "get_provider_fields": (".utils", "get_provider_fields"),
    "ModelResponseListIterator": (".utils", "ModelResponseListIterator"),
    "get_valid_models": (".utils", "get_valid_models"),
    "timeout": (".timeout", "timeout"),
    "get_llm_provider": (
        "litellm.litellm_core_utils.get_llm_provider_logic",
        "get_llm_provider",
    ),
    "remove_index_from_tool_calls": (
        "litellm.litellm_core_utils.core_helpers",
        "remove_index_from_tool_calls",
    ),
}

_COST_CALCULATOR_IMPORT_MAP = {
    "completion_cost": (".cost_calculator", "completion_cost"),
    "cost_per_token": (".cost_calculator", "cost_per_token"),
    "response_cost_calculator": (".cost_calculator", "response_cost_calculator"),
}

_TYPES_UTILS_IMPORT_MAP = {
    "ImageObject": (".types.utils", "ImageObject"),
    "BudgetConfig": (".types.utils", "BudgetConfig"),
    "all_litellm_params": (".types.utils", "all_litellm_params"),
    "_litellm_completion_params": (".types.utils", "all_litellm_params"),  # Alias
    "CredentialItem": (".types.utils", "CredentialItem"),
    "PriorityReservationDict": (".types.utils", "PriorityReservationDict"),
    "StandardKeyGenerationConfig": (".types.utils", "StandardKeyGenerationConfig"),
    "SearchProviders": (".types.utils", "SearchProviders"),
    "GenericStreamingChunk": (".types.utils", "GenericStreamingChunk"),
}

_TOKEN_COUNTER_IMPORT_MAP = {
    "get_modified_max_tokens": (
        "litellm.litellm_core_utils.token_counter",
        "get_modified_max_tokens",
    ),
}

_BEDROCK_TYPES_IMPORT_MAP = {
    "COHERE_EMBEDDING_INPUT_TYPES": (
        "litellm.types.llms.bedrock",
        "COHERE_EMBEDDING_INPUT_TYPES",
    ),
}

_CACHING_IMPORT_MAP = {
    "Cache": ("litellm.caching.caching", "Cache"),
    "DualCache": ("litellm.caching.caching", "DualCache"),
    "RedisCache": ("litellm.caching.caching", "RedisCache"),
    "InMemoryCache": ("litellm.caching.caching", "InMemoryCache"),
}

_LITELLM_LOGGING_IMPORT_MAP = {
    "Logging": ("litellm.litellm_core_utils.litellm_logging", "Logging"),
    "modify_integration": (
        "litellm.litellm_core_utils.litellm_logging",
        "modify_integration",
    ),
}

_DOTPROMPT_IMPORT_MAP = {
    "global_prompt_manager": (
        "litellm.integrations.dotprompt",
        "global_prompt_manager",
    ),
    "global_prompt_directory": (
        "litellm.integrations.dotprompt",
        "global_prompt_directory",
    ),
    "set_global_prompt_directory": (
        "litellm.integrations.dotprompt",
        "set_global_prompt_directory",
    ),
}

_TYPES_IMPORT_MAP = {
    "GuardrailItem": ("litellm.types.guardrails", "GuardrailItem"),
    "DefaultTeamSSOParams": (
        "litellm.types.proxy.management_endpoints.ui_sso",
        "DefaultTeamSSOParams",
    ),
    "LiteLLM_UpperboundKeyGenerateParams": (
        "litellm.types.proxy.management_endpoints.ui_sso",
        "LiteLLM_UpperboundKeyGenerateParams",
    ),
    "KeyManagementSystem": (
        "litellm.types.secret_managers.main",
        "KeyManagementSystem",
    ),
    "PriorityReservationSettings": (
        "litellm.types.utils",
        "PriorityReservationSettings",
    ),
    "CustomLogger": ("litellm.integrations.custom_logger", "CustomLogger"),
    "LoggingCallbackManager": (
        "litellm.litellm_core_utils.logging_callback_manager",
        "LoggingCallbackManager",
    ),
    "DatadogLLMObsInitParams": (
        "litellm.types.integrations.datadog_llm_obs",
        "DatadogLLMObsInitParams",
    ),
}

_LLM_PROVIDER_LOGIC_IMPORT_MAP = {
    "get_llm_provider": (
        "litellm.litellm_core_utils.get_llm_provider_logic",
        "get_llm_provider",
    ),
    "remove_index_from_tool_calls": (
        "litellm.litellm_core_utils.core_helpers",
        "remove_index_from_tool_calls",
    ),
}

_LLM_CONFIGS_IMPORT_MAP = {
    "AmazonConverseConfig": (
        ".llms.bedrock.chat.converse_transformation",
        "AmazonConverseConfig",
    ),
    "OpenAILikeChatConfig": (".llms.openai_like.chat.handler", "OpenAILikeChatConfig"),
    "GaladrielChatConfig": (
        ".llms.galadriel.chat.transformation",
        "GaladrielChatConfig",
    ),
    "GithubChatConfig": (".llms.github.chat.transformation", "GithubChatConfig"),
    "AzureAnthropicConfig": (
        ".llms.azure_ai.anthropic.transformation",
        "AzureAnthropicConfig",
    ),
    "BytezChatConfig": (".llms.bytez.chat.transformation", "BytezChatConfig"),
    "CompactifAIChatConfig": (
        ".llms.compactifai.chat.transformation",
        "CompactifAIChatConfig",
    ),
    "EmpowerChatConfig": (".llms.empower.chat.transformation", "EmpowerChatConfig"),
    "MinimaxChatConfig": (".llms.minimax.chat.transformation", "MinimaxChatConfig"),
    "AiohttpOpenAIChatConfig": (
        ".llms.aiohttp_openai.chat.transformation",
        "AiohttpOpenAIChatConfig",
    ),
    "HuggingFaceChatConfig": (
        ".llms.huggingface.chat.transformation",
        "HuggingFaceChatConfig",
    ),
    "HuggingFaceEmbeddingConfig": (
        ".llms.huggingface.embedding.transformation",
        "HuggingFaceEmbeddingConfig",
    ),
    "OobaboogaConfig": (".llms.oobabooga.chat.transformation", "OobaboogaConfig"),
    "MaritalkConfig": (".llms.maritalk", "MaritalkConfig"),
    "OpenrouterConfig": (".llms.openrouter.chat.transformation", "OpenrouterConfig"),
    "DataRobotConfig": (".llms.datarobot.chat.transformation", "DataRobotConfig"),
    "AnthropicConfig": (".llms.anthropic.chat.transformation", "AnthropicConfig"),
    "AnthropicTextConfig": (
        ".llms.anthropic.completion.transformation",
        "AnthropicTextConfig",
    ),
    "GroqSTTConfig": (".llms.groq.stt.transformation", "GroqSTTConfig"),
    "TritonConfig": (".llms.triton.completion.transformation", "TritonConfig"),
    "TritonGenerateConfig": (
        ".llms.triton.completion.transformation",
        "TritonGenerateConfig",
    ),
    "TritonInferConfig": (
        ".llms.triton.completion.transformation",
        "TritonInferConfig",
    ),
    "TritonEmbeddingConfig": (
        ".llms.triton.embedding.transformation",
        "TritonEmbeddingConfig",
    ),
    "HuggingFaceRerankConfig": (
        ".llms.huggingface.rerank.transformation",
        "HuggingFaceRerankConfig",
    ),
    "DatabricksConfig": (".llms.databricks.chat.transformation", "DatabricksConfig"),
    "DatabricksEmbeddingConfig": (
        ".llms.databricks.embed.transformation",
        "DatabricksEmbeddingConfig",
    ),
    "PredibaseConfig": (".llms.predibase.chat.transformation", "PredibaseConfig"),
    "ReplicateConfig": (".llms.replicate.chat.transformation", "ReplicateConfig"),
    "SnowflakeConfig": (".llms.snowflake.chat.transformation", "SnowflakeConfig"),
    "CohereRerankConfig": (".llms.cohere.rerank.transformation", "CohereRerankConfig"),
    "CohereRerankV2Config": (
        ".llms.cohere.rerank_v2.transformation",
        "CohereRerankV2Config",
    ),
    "AzureAIRerankConfig": (
        ".llms.azure_ai.rerank.transformation",
        "AzureAIRerankConfig",
    ),
    "InfinityRerankConfig": (
        ".llms.infinity.rerank.transformation",
        "InfinityRerankConfig",
    ),
    "JinaAIRerankConfig": (".llms.jina_ai.rerank.transformation", "JinaAIRerankConfig"),
    "DeepinfraRerankConfig": (
        ".llms.deepinfra.rerank.transformation",
        "DeepinfraRerankConfig",
    ),
    "HostedVLLMRerankConfig": (
        ".llms.hosted_vllm.rerank.transformation",
        "HostedVLLMRerankConfig",
    ),
    "NvidiaNimRerankConfig": (
        ".llms.nvidia_nim.rerank.transformation",
        "NvidiaNimRerankConfig",
    ),
    "NvidiaNimRankingConfig": (
        ".llms.nvidia_nim.rerank.ranking_transformation",
        "NvidiaNimRankingConfig",
    ),
    "VertexAIRerankConfig": (
        ".llms.vertex_ai.rerank.transformation",
        "VertexAIRerankConfig",
    ),
    "FireworksAIRerankConfig": (
        ".llms.fireworks_ai.rerank.transformation",
        "FireworksAIRerankConfig",
    ),
    "VoyageRerankConfig": (".llms.voyage.rerank.transformation", "VoyageRerankConfig"),
    "ClarifaiConfig": (".llms.clarifai.chat.transformation", "ClarifaiConfig"),
    "AI21ChatConfig": (".llms.ai21.chat.transformation", "AI21ChatConfig"),
    "LlamaAPIConfig": (".llms.meta_llama.chat.transformation", "LlamaAPIConfig"),
    "TogetherAITextCompletionConfig": (
        ".llms.together_ai.completion.transformation",
        "TogetherAITextCompletionConfig",
    ),
    "CloudflareChatConfig": (
        ".llms.cloudflare.chat.transformation",
        "CloudflareChatConfig",
    ),
    "NovitaConfig": (".llms.novita.chat.transformation", "NovitaConfig"),
    "PetalsConfig": (".llms.petals.completion.transformation", "PetalsConfig"),
    "OllamaChatConfig": (".llms.ollama.chat.transformation", "OllamaChatConfig"),
    "OllamaConfig": (".llms.ollama.completion.transformation", "OllamaConfig"),
    "SagemakerConfig": (".llms.sagemaker.completion.transformation", "SagemakerConfig"),
    "SagemakerChatConfig": (
        ".llms.sagemaker.chat.transformation",
        "SagemakerChatConfig",
    ),
    "CohereChatConfig": (".llms.cohere.chat.transformation", "CohereChatConfig"),
    "AnthropicMessagesConfig": (
        ".llms.anthropic.experimental_pass_through.messages.transformation",
        "AnthropicMessagesConfig",
    ),
    "AmazonAnthropicClaudeMessagesConfig": (
        ".llms.bedrock.messages.invoke_transformations.anthropic_claude3_transformation",
        "AmazonAnthropicClaudeMessagesConfig",
    ),
    "TogetherAIConfig": (".llms.together_ai.chat", "TogetherAIConfig"),
    "NLPCloudConfig": (".llms.nlp_cloud.chat.handler", "NLPCloudConfig"),
    "VertexGeminiConfig": (
        ".llms.vertex_ai.gemini.vertex_and_google_ai_studio_gemini",
        "VertexGeminiConfig",
    ),
    "GoogleAIStudioGeminiConfig": (
        ".llms.gemini.chat.transformation",
        "GoogleAIStudioGeminiConfig",
    ),
    "VertexAIAnthropicConfig": (
        ".llms.vertex_ai.vertex_ai_partner_models.anthropic.transformation",
        "VertexAIAnthropicConfig",
    ),
    "VertexAILlama3Config": (
        ".llms.vertex_ai.vertex_ai_partner_models.llama3.transformation",
        "VertexAILlama3Config",
    ),
    "VertexAIAi21Config": (
        ".llms.vertex_ai.vertex_ai_partner_models.ai21.transformation",
        "VertexAIAi21Config",
    ),
    "AmazonCohereChatConfig": (
        ".llms.bedrock.chat.invoke_handler",
        "AmazonCohereChatConfig",
    ),
    "AmazonBedrockGlobalConfig": (
        ".llms.bedrock.common_utils",
        "AmazonBedrockGlobalConfig",
    ),
    "AmazonAI21Config": (
        ".llms.bedrock.chat.invoke_transformations.amazon_ai21_transformation",
        "AmazonAI21Config",
    ),
    "AmazonInvokeNovaConfig": (
        ".llms.bedrock.chat.invoke_transformations.amazon_nova_transformation",
        "AmazonInvokeNovaConfig",
    ),
    "AmazonQwen2Config": (
        ".llms.bedrock.chat.invoke_transformations.amazon_qwen2_transformation",
        "AmazonQwen2Config",
    ),
    "AmazonQwen3Config": (
        ".llms.bedrock.chat.invoke_transformations.amazon_qwen3_transformation",
        "AmazonQwen3Config",
    ),
    # Aliases for backwards compatibility
    "VertexAIConfig": (
        ".llms.vertex_ai.gemini.vertex_and_google_ai_studio_gemini",
        "VertexGeminiConfig",
    ),  # Alias
    "GeminiConfig": (
        ".llms.gemini.chat.transformation",
        "GoogleAIStudioGeminiConfig",
    ),  # Alias
    "AmazonAnthropicConfig": (
        ".llms.bedrock.chat.invoke_transformations.anthropic_claude2_transformation",
        "AmazonAnthropicConfig",
    ),
    "AmazonAnthropicClaudeConfig": (
        ".llms.bedrock.chat.invoke_transformations.anthropic_claude3_transformation",
        "AmazonAnthropicClaudeConfig",
    ),
    "AmazonCohereConfig": (
        ".llms.bedrock.chat.invoke_transformations.amazon_cohere_transformation",
        "AmazonCohereConfig",
    ),
    "AmazonLlamaConfig": (
        ".llms.bedrock.chat.invoke_transformations.amazon_llama_transformation",
        "AmazonLlamaConfig",
    ),
    "AmazonDeepSeekR1Config": (
        ".llms.bedrock.chat.invoke_transformations.amazon_deepseek_transformation",
        "AmazonDeepSeekR1Config",
    ),
    "AmazonMistralConfig": (
        ".llms.bedrock.chat.invoke_transformations.amazon_mistral_transformation",
        "AmazonMistralConfig",
    ),
    "AmazonMoonshotConfig": (
        ".llms.bedrock.chat.invoke_transformations.amazon_moonshot_transformation",
        "AmazonMoonshotConfig",
    ),
    "AmazonTitanConfig": (
        ".llms.bedrock.chat.invoke_transformations.amazon_titan_transformation",
        "AmazonTitanConfig",
    ),
    "AmazonTwelveLabsPegasusConfig": (
        ".llms.bedrock.chat.invoke_transformations.amazon_twelvelabs_pegasus_transformation",
        "AmazonTwelveLabsPegasusConfig",
    ),
    "AmazonInvokeConfig": (
        ".llms.bedrock.chat.invoke_transformations.base_invoke_transformation",
        "AmazonInvokeConfig",
    ),
    "AmazonBedrockOpenAIConfig": (
        ".llms.bedrock.chat.invoke_transformations.amazon_openai_transformation",
        "AmazonBedrockOpenAIConfig",
    ),
    "AmazonStabilityConfig": (
        ".llms.bedrock.image_generation.amazon_stability1_transformation",
        "AmazonStabilityConfig",
    ),
    "AmazonStability3Config": (
        ".llms.bedrock.image_generation.amazon_stability3_transformation",
        "AmazonStability3Config",
    ),
    "AmazonNovaCanvasConfig": (
        ".llms.bedrock.image_generation.amazon_nova_canvas_transformation",
        "AmazonNovaCanvasConfig",
    ),
    "AmazonTitanG1Config": (
        ".llms.bedrock.embed.amazon_titan_g1_transformation",
        "AmazonTitanG1Config",
    ),
    "AmazonTitanMultimodalEmbeddingG1Config": (
        ".llms.bedrock.embed.amazon_titan_multimodal_transformation",
        "AmazonTitanMultimodalEmbeddingG1Config",
    ),
    "CohereV2ChatConfig": (".llms.cohere.chat.v2_transformation", "CohereV2ChatConfig"),
    "BedrockCohereEmbeddingConfig": (
        ".llms.bedrock.embed.cohere_transformation",
        "BedrockCohereEmbeddingConfig",
    ),
    "TwelveLabsMarengoEmbeddingConfig": (
        ".llms.bedrock.embed.twelvelabs_marengo_transformation",
        "TwelveLabsMarengoEmbeddingConfig",
    ),
    "AmazonNovaEmbeddingConfig": (
        ".llms.bedrock.embed.amazon_nova_transformation",
        "AmazonNovaEmbeddingConfig",
    ),
    "OpenAIConfig": (".llms.openai.openai", "OpenAIConfig"),
    "MistralEmbeddingConfig": (".llms.openai.openai", "MistralEmbeddingConfig"),
    "OpenAIImageVariationConfig": (
        ".llms.openai.image_variations.transformation",
        "OpenAIImageVariationConfig",
    ),
    "DeepInfraConfig": (".llms.deepinfra.chat.transformation", "DeepInfraConfig"),
    "DeepgramAudioTranscriptionConfig": (
        ".llms.deepgram.audio_transcription.transformation",
        "DeepgramAudioTranscriptionConfig",
    ),
    "TopazImageVariationConfig": (
        ".llms.topaz.image_variations.transformation",
        "TopazImageVariationConfig",
    ),
    "OpenAITextCompletionConfig": (
        "litellm.llms.openai.completion.transformation",
        "OpenAITextCompletionConfig",
    ),
    "GroqChatConfig": (".llms.groq.chat.transformation", "GroqChatConfig"),
    "A2AConfig": (".llms.a2a.chat.transformation", "A2AConfig"),
    "GenAIHubOrchestrationConfig": (
        ".llms.sap.chat.transformation",
        "GenAIHubOrchestrationConfig",
    ),
    "VoyageEmbeddingConfig": (
        ".llms.voyage.embedding.transformation",
        "VoyageEmbeddingConfig",
    ),
    "VoyageContextualEmbeddingConfig": (
        ".llms.voyage.embedding.transformation_contextual",
        "VoyageContextualEmbeddingConfig",
    ),
    "InfinityEmbeddingConfig": (
        ".llms.infinity.embedding.transformation",
        "InfinityEmbeddingConfig",
    ),
    "AzureAIStudioConfig": (
        ".llms.azure_ai.chat.transformation",
        "AzureAIStudioConfig",
    ),
    "MistralConfig": (".llms.mistral.chat.transformation", "MistralConfig"),
    "OpenAIResponsesAPIConfig": (
        ".llms.openai.responses.transformation",
        "OpenAIResponsesAPIConfig",
    ),
    "AzureOpenAIResponsesAPIConfig": (
        ".llms.azure.responses.transformation",
        "AzureOpenAIResponsesAPIConfig",
    ),
    "AzureOpenAIOSeriesResponsesAPIConfig": (
        ".llms.azure.responses.o_series_transformation",
        "AzureOpenAIOSeriesResponsesAPIConfig",
    ),
    "XAIResponsesAPIConfig": (
        ".llms.xai.responses.transformation",
        "XAIResponsesAPIConfig",
    ),
    "LiteLLMProxyResponsesAPIConfig": (
        ".llms.litellm_proxy.responses.transformation",
        "LiteLLMProxyResponsesAPIConfig",
    ),
    "VolcEngineResponsesAPIConfig": (
        ".llms.volcengine.responses.transformation",
        "VolcEngineResponsesAPIConfig",
    ),
    "ManusResponsesAPIConfig": (
        ".llms.manus.responses.transformation",
        "ManusResponsesAPIConfig",
    ),
    "PerplexityResponsesConfig": (
        ".llms.perplexity.responses.transformation",
        "PerplexityResponsesConfig",
    ),
    "GoogleAIStudioInteractionsConfig": (
        ".llms.gemini.interactions.transformation",
        "GoogleAIStudioInteractionsConfig",
    ),
    "OpenAIOSeriesConfig": (
        ".llms.openai.chat.o_series_transformation",
        "OpenAIOSeriesConfig",
    ),
    "AnthropicSkillsConfig": (
        ".llms.anthropic.skills.transformation",
        "AnthropicSkillsConfig",
    ),
    "BaseSkillsAPIConfig": (
        ".llms.base_llm.skills.transformation",
        "BaseSkillsAPIConfig",
    ),
    "GradientAIConfig": (".llms.gradient_ai.chat.transformation", "GradientAIConfig"),
    # Alias for backwards compatibility
    "OpenAIO1Config": (
        ".llms.openai.chat.o_series_transformation",
        "OpenAIOSeriesConfig",
    ),  # Alias
    "OpenAIGPTConfig": (".llms.openai.chat.gpt_transformation", "OpenAIGPTConfig"),
    "OpenAIGPT5Config": (".llms.openai.chat.gpt_5_transformation", "OpenAIGPT5Config"),
    "OpenAIWhisperAudioTranscriptionConfig": (
        ".llms.openai.transcriptions.whisper_transformation",
        "OpenAIWhisperAudioTranscriptionConfig",
    ),
    "OpenAIGPTAudioTranscriptionConfig": (
        ".llms.openai.transcriptions.gpt_transformation",
        "OpenAIGPTAudioTranscriptionConfig",
    ),
    "OpenAIGPTAudioConfig": (
        ".llms.openai.chat.gpt_audio_transformation",
        "OpenAIGPTAudioConfig",
    ),
    "NvidiaNimConfig": (".llms.nvidia_nim.chat.transformation", "NvidiaNimConfig"),
    "NvidiaNimEmbeddingConfig": (".llms.nvidia_nim.embed", "NvidiaNimEmbeddingConfig"),
    "FeatherlessAIConfig": (
        ".llms.featherless_ai.chat.transformation",
        "FeatherlessAIConfig",
    ),
    "CerebrasConfig": (".llms.cerebras.chat", "CerebrasConfig"),
    "BasetenConfig": (".llms.baseten.chat", "BasetenConfig"),
    "SambanovaConfig": (".llms.sambanova.chat", "SambanovaConfig"),
    "SambaNovaEmbeddingConfig": (
        ".llms.sambanova.embedding.transformation",
        "SambaNovaEmbeddingConfig",
    ),
    "FireworksAIConfig": (
        ".llms.fireworks_ai.chat.transformation",
        "FireworksAIConfig",
    ),
    "FireworksAITextCompletionConfig": (
        ".llms.fireworks_ai.completion.transformation",
        "FireworksAITextCompletionConfig",
    ),
    "FireworksAIAudioTranscriptionConfig": (
        ".llms.fireworks_ai.audio_transcription.transformation",
        "FireworksAIAudioTranscriptionConfig",
    ),
    "FireworksAIEmbeddingConfig": (
        ".llms.fireworks_ai.embed.fireworks_ai_transformation",
        "FireworksAIEmbeddingConfig",
    ),
    "FriendliaiChatConfig": (
        ".llms.friendliai.chat.transformation",
        "FriendliaiChatConfig",
    ),
    "JinaAIEmbeddingConfig": (
        ".llms.jina_ai.embedding.transformation",
        "JinaAIEmbeddingConfig",
    ),
    "XAIChatConfig": (".llms.xai.chat.transformation", "XAIChatConfig"),
    "ZAIChatConfig": (".llms.zai.chat.transformation", "ZAIChatConfig"),
    "AIMLChatConfig": (".llms.aiml.chat.transformation", "AIMLChatConfig"),
    "VolcEngineChatConfig": (
        ".llms.volcengine.chat.transformation",
        "VolcEngineChatConfig",
    ),
    "CodestralTextCompletionConfig": (
        ".llms.codestral.completion.transformation",
        "CodestralTextCompletionConfig",
    ),
    "AzureOpenAIAssistantsAPIConfig": (
        ".llms.azure.azure",
        "AzureOpenAIAssistantsAPIConfig",
    ),
    "HerokuChatConfig": (".llms.heroku.chat.transformation", "HerokuChatConfig"),
    "CometAPIConfig": (".llms.cometapi.chat.transformation", "CometAPIConfig"),
    "AzureOpenAIConfig": (".llms.azure.chat.gpt_transformation", "AzureOpenAIConfig"),
    "AzureOpenAIGPT5Config": (
        ".llms.azure.chat.gpt_5_transformation",
        "AzureOpenAIGPT5Config",
    ),
    "AzureOpenAITextConfig": (
        ".llms.azure.completion.transformation",
        "AzureOpenAITextConfig",
    ),
    "HostedVLLMChatConfig": (
        ".llms.hosted_vllm.chat.transformation",
        "HostedVLLMChatConfig",
    ),
    "HostedVLLMEmbeddingConfig": (
        ".llms.hosted_vllm.embedding.transformation",
        "HostedVLLMEmbeddingConfig",
    ),
    # Alias for backwards compatibility
    "VolcEngineConfig": (
        ".llms.volcengine.chat.transformation",
        "VolcEngineChatConfig",
    ),  # Alias
    "LlamafileChatConfig": (
        ".llms.llamafile.chat.transformation",
        "LlamafileChatConfig",
    ),
    "LiteLLMProxyChatConfig": (
        ".llms.litellm_proxy.chat.transformation",
        "LiteLLMProxyChatConfig",
    ),
    "VLLMConfig": (".llms.vllm.completion.transformation", "VLLMConfig"),
    "DeepSeekChatConfig": (".llms.deepseek.chat.transformation", "DeepSeekChatConfig"),
    "LMStudioChatConfig": (".llms.lm_studio.chat.transformation", "LMStudioChatConfig"),
    "LmStudioEmbeddingConfig": (
        ".llms.lm_studio.embed.transformation",
        "LmStudioEmbeddingConfig",
    ),
    "NscaleConfig": (".llms.nscale.chat.transformation", "NscaleConfig"),
    "PerplexityChatConfig": (
        ".llms.perplexity.chat.transformation",
        "PerplexityChatConfig",
    ),
    "AzureOpenAIO1Config": (
        ".llms.azure.chat.o_series_transformation",
        "AzureOpenAIO1Config",
    ),
    "IBMWatsonXAIConfig": (
        ".llms.watsonx.completion.transformation",
        "IBMWatsonXAIConfig",
    ),
    "IBMWatsonXChatConfig": (
        ".llms.watsonx.chat.transformation",
        "IBMWatsonXChatConfig",
    ),
    "IBMWatsonXEmbeddingConfig": (
        ".llms.watsonx.embed.transformation",
        "IBMWatsonXEmbeddingConfig",
    ),
    "GenAIHubEmbeddingConfig": (
        ".llms.sap.embed.transformation",
        "GenAIHubEmbeddingConfig",
    ),
    "IBMWatsonXAudioTranscriptionConfig": (
        ".llms.watsonx.audio_transcription.transformation",
        "IBMWatsonXAudioTranscriptionConfig",
    ),
    "GithubCopilotConfig": (
        ".llms.github_copilot.chat.transformation",
        "GithubCopilotConfig",
    ),
    "GithubCopilotResponsesAPIConfig": (
        ".llms.github_copilot.responses.transformation",
        "GithubCopilotResponsesAPIConfig",
    ),
    "GithubCopilotEmbeddingConfig": (
        ".llms.github_copilot.embedding.transformation",
        "GithubCopilotEmbeddingConfig",
    ),
    "ChatGPTConfig": (".llms.chatgpt.chat.transformation", "ChatGPTConfig"),
    "ChatGPTResponsesAPIConfig": (
        ".llms.chatgpt.responses.transformation",
        "ChatGPTResponsesAPIConfig",
    ),
    "NebiusConfig": (".llms.nebius.chat.transformation", "NebiusConfig"),
    "WandbConfig": (".llms.wandb.chat.transformation", "WandbConfig"),
    "GigaChatConfig": (".llms.gigachat.chat.transformation", "GigaChatConfig"),
    "GigaChatEmbeddingConfig": (
        ".llms.gigachat.embedding.transformation",
        "GigaChatEmbeddingConfig",
    ),
    "DashScopeChatConfig": (
        ".llms.dashscope.chat.transformation",
        "DashScopeChatConfig",
    ),
    "MoonshotChatConfig": (".llms.moonshot.chat.transformation", "MoonshotChatConfig"),
    "DockerModelRunnerChatConfig": (
        ".llms.docker_model_runner.chat.transformation",
        "DockerModelRunnerChatConfig",
    ),
    "V0ChatConfig": (".llms.v0.chat.transformation", "V0ChatConfig"),
    "OCIChatConfig": (".llms.oci.chat.transformation", "OCIChatConfig"),
    "MorphChatConfig": (".llms.morph.chat.transformation", "MorphChatConfig"),
    "RAGFlowConfig": (".llms.ragflow.chat.transformation", "RAGFlowConfig"),
    "LambdaAIChatConfig": (".llms.lambda_ai.chat.transformation", "LambdaAIChatConfig"),
    "HyperbolicChatConfig": (
        ".llms.hyperbolic.chat.transformation",
        "HyperbolicChatConfig",
    ),
    "VercelAIGatewayConfig": (
        ".llms.vercel_ai_gateway.chat.transformation",
        "VercelAIGatewayConfig",
    ),
    "OVHCloudChatConfig": (".llms.ovhcloud.chat.transformation", "OVHCloudChatConfig"),
    "OVHCloudEmbeddingConfig": (
        ".llms.ovhcloud.embedding.transformation",
        "OVHCloudEmbeddingConfig",
    ),
    "CometAPIEmbeddingConfig": (
        ".llms.cometapi.embed.transformation",
        "CometAPIEmbeddingConfig",
    ),
    "LemonadeChatConfig": (".llms.lemonade.chat.transformation", "LemonadeChatConfig"),
    "SnowflakeEmbeddingConfig": (
        ".llms.snowflake.embedding.transformation",
        "SnowflakeEmbeddingConfig",
    ),
    "AmazonNovaChatConfig": (
        ".llms.amazon_nova.chat.transformation",
        "AmazonNovaChatConfig",
    ),
}

# Import map for utils module lazy imports
_UTILS_MODULE_IMPORT_MAP = {
    "encoding": ("litellm.main", "encoding"),
    "BaseVectorStore": (
        "litellm.integrations.vector_store_integrations.base_vector_store",
        "BaseVectorStore",
    ),
    "CredentialAccessor": (
        "litellm.litellm_core_utils.credential_accessor",
        "CredentialAccessor",
    ),
    "exception_type": (
        "litellm.litellm_core_utils.exception_mapping_utils",
        "exception_type",
    ),
    "get_error_message": (
        "litellm.litellm_core_utils.exception_mapping_utils",
        "get_error_message",
    ),
    "_get_response_headers": (
        "litellm.litellm_core_utils.exception_mapping_utils",
        "_get_response_headers",
    ),
    "get_llm_provider": (
        "litellm.litellm_core_utils.get_llm_provider_logic",
        "get_llm_provider",
    ),
    "_is_non_openai_azure_model": (
        "litellm.litellm_core_utils.get_llm_provider_logic",
        "_is_non_openai_azure_model",
    ),
    "get_supported_openai_params": (
        "litellm.litellm_core_utils.get_supported_openai_params",
        "get_supported_openai_params",
    ),
    "LiteLLMResponseObjectHandler": (
        "litellm.litellm_core_utils.llm_response_utils.convert_dict_to_response",
        "LiteLLMResponseObjectHandler",
    ),
    "_handle_invalid_parallel_tool_calls": (
        "litellm.litellm_core_utils.llm_response_utils.convert_dict_to_response",
        "_handle_invalid_parallel_tool_calls",
    ),
    "convert_to_model_response_object": (
        "litellm.litellm_core_utils.llm_response_utils.convert_dict_to_response",
        "convert_to_model_response_object",
    ),
    "convert_to_streaming_response": (
        "litellm.litellm_core_utils.llm_response_utils.convert_dict_to_response",
        "convert_to_streaming_response",
    ),
    "convert_to_streaming_response_async": (
        "litellm.litellm_core_utils.llm_response_utils.convert_dict_to_response",
        "convert_to_streaming_response_async",
    ),
    "get_api_base": (
        "litellm.litellm_core_utils.llm_response_utils.get_api_base",
        "get_api_base",
    ),
    "ResponseMetadata": (
        "litellm.litellm_core_utils.llm_response_utils.response_metadata",
        "ResponseMetadata",
    ),
    "_parse_content_for_reasoning": (
        "litellm.litellm_core_utils.prompt_templates.common_utils",
        "_parse_content_for_reasoning",
    ),
    "LiteLLMLoggingObject": (
        "litellm.litellm_core_utils.redact_messages",
        "LiteLLMLoggingObject",
    ),
    "redact_message_input_output_from_logging": (
        "litellm.litellm_core_utils.redact_messages",
        "redact_message_input_output_from_logging",
    ),
    "CustomStreamWrapper": (
        "litellm.litellm_core_utils.streaming_handler",
        "CustomStreamWrapper",
    ),
    "BaseGoogleGenAIGenerateContentConfig": (
        "litellm.llms.base_llm.google_genai.transformation",
        "BaseGoogleGenAIGenerateContentConfig",
    ),
    "BaseOCRConfig": ("litellm.llms.base_llm.ocr.transformation", "BaseOCRConfig"),
    "BaseSearchConfig": (
        "litellm.llms.base_llm.search.transformation",
        "BaseSearchConfig",
    ),
    "BaseTextToSpeechConfig": (
        "litellm.llms.base_llm.text_to_speech.transformation",
        "BaseTextToSpeechConfig",
    ),
    "BedrockModelInfo": ("litellm.llms.bedrock.common_utils", "BedrockModelInfo"),
    "CohereModelInfo": ("litellm.llms.cohere.common_utils", "CohereModelInfo"),
    "MistralOCRConfig": ("litellm.llms.mistral.ocr.transformation", "MistralOCRConfig"),
    "Rules": ("litellm.litellm_core_utils.rules", "Rules"),
    "AsyncHTTPHandler": ("litellm.llms.custom_httpx.http_handler", "AsyncHTTPHandler"),
    "HTTPHandler": ("litellm.llms.custom_httpx.http_handler", "HTTPHandler"),
    "get_num_retries_from_retry_policy": (
        "litellm.router_utils.get_retry_from_policy",
        "get_num_retries_from_retry_policy",
    ),
    "reset_retry_policy": (
        "litellm.router_utils.get_retry_from_policy",
        "reset_retry_policy",
    ),
    "get_secret": ("litellm.secret_managers.main", "get_secret"),
    "get_coroutine_checker": (
        "litellm.litellm_core_utils.cached_imports",
        "get_coroutine_checker",
    ),
    "get_litellm_logging_class": (
        "litellm.litellm_core_utils.cached_imports",
        "get_litellm_logging_class",
    ),
    "get_set_callbacks": (
        "litellm.litellm_core_utils.cached_imports",
        "get_set_callbacks",
    ),
    "get_litellm_metadata_from_kwargs": (
        "litellm.litellm_core_utils.core_helpers",
        "get_litellm_metadata_from_kwargs",
    ),
    "map_finish_reason": (
        "litellm.litellm_core_utils.core_helpers",
        "map_finish_reason",
    ),
    "process_response_headers": (
        "litellm.litellm_core_utils.core_helpers",
        "process_response_headers",
    ),
    "delete_nested_value": (
        "litellm.litellm_core_utils.dot_notation_indexing",
        "delete_nested_value",
    ),
    "is_nested_path": (
        "litellm.litellm_core_utils.dot_notation_indexing",
        "is_nested_path",
    ),
    "_get_base_model_from_litellm_call_metadata": (
        "litellm.litellm_core_utils.get_litellm_params",
        "_get_base_model_from_litellm_call_metadata",
    ),
    "get_litellm_params": (
        "litellm.litellm_core_utils.get_litellm_params",
        "get_litellm_params",
    ),
    "_ensure_extra_body_is_safe": (
        "litellm.litellm_core_utils.llm_request_utils",
        "_ensure_extra_body_is_safe",
    ),
    "get_formatted_prompt": (
        "litellm.litellm_core_utils.llm_response_utils.get_formatted_prompt",
        "get_formatted_prompt",
    ),
    "get_response_headers": (
        "litellm.litellm_core_utils.llm_response_utils.get_headers",
        "get_response_headers",
    ),
    "update_response_metadata": (
        "litellm.litellm_core_utils.llm_response_utils.response_metadata",
        "update_response_metadata",
    ),
    "executor": ("litellm.litellm_core_utils.thread_pool_executor", "executor"),
    "BaseAnthropicMessagesConfig": (
        "litellm.llms.base_llm.anthropic_messages.transformation",
        "BaseAnthropicMessagesConfig",
    ),
    "BaseAudioTranscriptionConfig": (
        "litellm.llms.base_llm.audio_transcription.transformation",
        "BaseAudioTranscriptionConfig",
    ),
    "BaseBatchesConfig": (
        "litellm.llms.base_llm.batches.transformation",
        "BaseBatchesConfig",
    ),
    "BaseContainerConfig": (
        "litellm.llms.base_llm.containers.transformation",
        "BaseContainerConfig",
    ),
    "BaseEmbeddingConfig": (
        "litellm.llms.base_llm.embedding.transformation",
        "BaseEmbeddingConfig",
    ),
    "BaseImageEditConfig": (
        "litellm.llms.base_llm.image_edit.transformation",
        "BaseImageEditConfig",
    ),
    "BaseImageGenerationConfig": (
        "litellm.llms.base_llm.image_generation.transformation",
        "BaseImageGenerationConfig",
    ),
    "BaseImageVariationConfig": (
        "litellm.llms.base_llm.image_variations.transformation",
        "BaseImageVariationConfig",
    ),
    "BasePassthroughConfig": (
        "litellm.llms.base_llm.passthrough.transformation",
        "BasePassthroughConfig",
    ),
    "BaseRealtimeConfig": (
        "litellm.llms.base_llm.realtime.transformation",
        "BaseRealtimeConfig",
    ),
    "BaseRerankConfig": (
        "litellm.llms.base_llm.rerank.transformation",
        "BaseRerankConfig",
    ),
    "BaseVectorStoreConfig": (
        "litellm.llms.base_llm.vector_store.transformation",
        "BaseVectorStoreConfig",
    ),
    "BaseVectorStoreFilesConfig": (
        "litellm.llms.base_llm.vector_store_files.transformation",
        "BaseVectorStoreFilesConfig",
    ),
    "BaseVideoConfig": (
        "litellm.llms.base_llm.videos.transformation",
        "BaseVideoConfig",
    ),
    "ANTHROPIC_API_ONLY_HEADERS": (
        "litellm.types.llms.anthropic",
        "ANTHROPIC_API_ONLY_HEADERS",
    ),
    "AnthropicThinkingParam": (
        "litellm.types.llms.anthropic",
        "AnthropicThinkingParam",
    ),
    "RerankResponse": ("litellm.types.rerank", "RerankResponse"),
    "ChatCompletionDeltaToolCallChunk": (
        "litellm.types.llms.openai",
        "ChatCompletionDeltaToolCallChunk",
    ),
    "ChatCompletionToolCallChunk": (
        "litellm.types.llms.openai",
        "ChatCompletionToolCallChunk",
    ),
    "ChatCompletionToolCallFunctionChunk": (
        "litellm.types.llms.openai",
        "ChatCompletionToolCallFunctionChunk",
    ),
    "LiteLLM_Params": ("litellm.types.router", "LiteLLM_Params"),
}

# Export all name tuples and import maps for use in _lazy_imports.py
__all__ = [
    # Name tuples
    "COST_CALCULATOR_NAMES",
    "LITELLM_LOGGING_NAMES",
    "UTILS_NAMES",
    "TOKEN_COUNTER_NAMES",
    "LLM_CLIENT_CACHE_NAMES",
    "BEDROCK_TYPES_NAMES",
    "TYPES_UTILS_NAMES",
    "CACHING_NAMES",
    "HTTP_HANDLER_NAMES",
    "DOTPROMPT_NAMES",
    "LLM_CONFIG_NAMES",
    "TYPES_NAMES",
    "LLM_PROVIDER_LOGIC_NAMES",
    "UTILS_MODULE_NAMES",
    # Import maps
    "_UTILS_IMPORT_MAP",
    "_COST_CALCULATOR_IMPORT_MAP",
    "_TYPES_UTILS_IMPORT_MAP",
    "_TOKEN_COUNTER_IMPORT_MAP",
    "_BEDROCK_TYPES_IMPORT_MAP",
    "_CACHING_IMPORT_MAP",
    "_LITELLM_LOGGING_IMPORT_MAP",
    "_DOTPROMPT_IMPORT_MAP",
    "_TYPES_IMPORT_MAP",
    "_LLM_CONFIGS_IMPORT_MAP",
    "_LLM_PROVIDER_LOGIC_IMPORT_MAP",
    "_UTILS_MODULE_IMPORT_MAP",
]



from enum import Enum
from typing import List, Optional

from typing_extensions import NotRequired, TypedDict


class WatsonXAPIParams(TypedDict):
    project_id: Optional[str]
    space_id: Optional[str]
    region_name: Optional[str]


class WatsonXCredentials(TypedDict):
    api_key: str
    api_base: str
    token: Optional[str]


class WatsonXAudioTranscriptionRequestBody(TypedDict):
    """
    WatsonX Audio Transcription API request body.

    Follows multipart/form-data format for WatsonX Whisper models.
    See: https://cloud.ibm.com/apidocs/watsonx-ai
    """

    model: str
    """Model name (e.g., 'whisper-large-v3-turbo')"""

    project_id: NotRequired[str]
    """WatsonX project ID (optional)"""

    space_id: NotRequired[str]
    """WatsonX space ID (optional)"""

    language: NotRequired[str]
    """Language code (e.g., 'en', 'es')"""

    prompt: NotRequired[str]
    """Optional prompt to guide transcription"""

    response_format: NotRequired[str]
    """Response format: 'json', 'text', 'srt', 'verbose_json', 'vtt'"""

    temperature: NotRequired[float]
    """Sampling temperature (0-1)"""

    timestamp_granularities: NotRequired[List[str]]
    """Timestamp granularities: ['word', 'segment']"""


class WatsonXAIEndpoint(str, Enum):
    TEXT_GENERATION = "/ml/v1/text/generation"
    TEXT_GENERATION_STREAM = "/ml/v1/text/generation_stream"
    CHAT = "/ml/v1/text/chat"
    CHAT_STREAM = "/ml/v1/text/chat_stream"
    DEPLOYMENT_TEXT_GENERATION = "/ml/v1/deployments/{deployment_id}/text/generation"
    DEPLOYMENT_TEXT_GENERATION_STREAM = (
        "/ml/v1/deployments/{deployment_id}/text/generation_stream"
    )
    DEPLOYMENT_CHAT = "/ml/v1/deployments/{deployment_id}/text/chat"
    DEPLOYMENT_CHAT_STREAM = "/ml/v1/deployments/{deployment_id}/text/chat_stream"
    EMBEDDINGS = "/ml/v1/text/embeddings"
    PROMPTS = "/ml/v1/prompts"
    AVAILABLE_MODELS = "/ml/v1/foundation_model_specs"


class WatsonXModelPattern(str, Enum):
    """Model identifier patterns for WatsonX models"""

    GRANITE_CHAT = "granite-chat"
    IBM_MISTRAL = "ibm-mistral"
    IBM_MISTRALAI = "ibm-mistralai"
    GPT_OSS = "openai/gpt-oss"
    LLAMA3_INSTRUCT = "meta-llama/llama-3-instruct"



"""
Translate from OpenAI's `/v1/chat/completions` to Amazon Nova's `/v1/chat/completions`
"""
from typing import Any, List, Optional, Tuple

import httpx

import litellm
from litellm.litellm_core_utils.litellm_logging import Logging as LiteLLMLoggingObj
from litellm.secret_managers.main import get_secret_str
from litellm.types.llms.openai import (
    AllMessageValues,
)
from litellm.types.utils import ModelResponse

from ...openai_like.chat.transformation import OpenAILikeChatConfig


class AmazonNovaChatConfig(OpenAILikeChatConfig):
    max_completion_tokens: Optional[int] = None
    max_tokens: Optional[int] = None
    metadata: Optional[int] = None
    temperature: Optional[int] = None
    top_p: Optional[int] = None
    tools: Optional[list] = None
    reasoning_effort: Optional[list] = None

    def __init__(
        self,
        max_completion_tokens: Optional[int] = None,
        max_tokens: Optional[int] = None,
        temperature: Optional[int] = None,
        top_p: Optional[int] = None,
        tools: Optional[list] = None,
        reasoning_effort: Optional[list] = None,
    ) -> None:
        locals_ = locals().copy()
        for key, value in locals_.items():
            if key != "self" and value is not None:
                setattr(self.__class__, key, value)

    @property
    def custom_llm_provider(self) -> Optional[str]:
        return "amazon_nova"

    @classmethod
    def get_config(cls):
        return super().get_config()

    def _get_openai_compatible_provider_info(
        self, api_base: Optional[str], api_key: Optional[str]
    ) -> Tuple[Optional[str], Optional[str]]:
        # Amazon Nova is openai compatible, we just need to set this to custom_openai and have the api_base be Nova's endpoint
        api_base = (
            api_base
            or get_secret_str("AMAZON_NOVA_API_BASE")
            or "https://api.nova.amazon.com/v1"
        )  # type: ignore
        
        # Get API key from multiple sources
        key = (
            api_key
            or litellm.amazon_nova_api_key
            or get_secret_str("AMAZON_NOVA_API_KEY")
            or litellm.api_key
        )
        return api_base, key
    
    def get_supported_openai_params(self, model: str) -> List:
        return [
            "top_p",
            "temperature",
            "max_tokens",
            "max_completion_tokens",
            "metadata",
            "stop",
            "stream",
            "stream_options",
            "tools",
            "tool_choice",
            "reasoning_effort"
        ]

    def transform_response(
        self,
        model: str,
        raw_response: httpx.Response,
        model_response: ModelResponse,
        logging_obj: LiteLLMLoggingObj,
        request_data: dict,
        messages: List[AllMessageValues],
        optional_params: dict,
        litellm_params: dict,
        encoding: Any,
        api_key: Optional[str] = None,
        json_mode: Optional[bool] = None,
    ) -> ModelResponse:
        model_response = super().transform_response(
            model=model,
            model_response=model_response,
            raw_response=raw_response,
            messages=messages,
            logging_obj=logging_obj,
            request_data=request_data,
            encoding=encoding,
            optional_params=optional_params,
            json_mode=json_mode,
            litellm_params=litellm_params,
            api_key=api_key,
        )

        # Storing amazon_nova in the model response for easier cost calculation later
        setattr(model_response, "model", "amazon-nova/" + model)

        return model_response


"""
Transformation for Azure AI Foundry Model Router.

The Model Router is a special Azure AI deployment that automatically routes requests
to the best available model. It has specific cost tracking requirements.
"""
from typing import Any, List, Optional

from httpx import Response

from litellm.llms.azure_ai.chat.transformation import AzureAIStudioConfig
from litellm.llms.base_llm.chat.transformation import LiteLLMLoggingObj
from litellm.types.llms.openai import AllMessageValues
from litellm.types.utils import ModelResponse


class AzureModelRouterConfig(AzureAIStudioConfig):
    """
    Configuration for Azure AI Foundry Model Router.
    
    Handles:
    - Stripping model_router prefix before sending to Azure API
    - Preserving full model path in responses for cost tracking
    - Calculating flat infrastructure costs for Model Router
    """

    def transform_request(
        self,
        model: str,
        messages: List[AllMessageValues],
        optional_params: dict,
        litellm_params: dict,
        headers: dict,
    ) -> dict:
        """
        Transform request for Model Router.
        
        Strips the model_router/ prefix so only the deployment name is sent to Azure.
        Example: model_router/azure-model-router -> azure-model-router
        """
        from litellm.llms.azure_ai.common_utils import AzureFoundryModelInfo

        # Get base model name (strips routing prefixes like model_router/)
        base_model: str = AzureFoundryModelInfo.get_base_model(model)
        
        return super().transform_request(
            base_model, messages, optional_params, litellm_params, headers
        )

    def transform_response(
        self,
        model: str,
        raw_response: Response,
        model_response: ModelResponse,
        logging_obj: LiteLLMLoggingObj,
        request_data: dict,
        messages: List[AllMessageValues],
        optional_params: dict,
        litellm_params: dict,
        encoding: Any,
        api_key: Optional[str] = None,
        json_mode: Optional[bool] = None,
    ) -> ModelResponse:
        """
        Transform response for Model Router.
        
        Preserves the original model path (including model_router/ prefix) in the response
        for proper cost tracking and logging.
        """
        from litellm.llms.azure_ai.common_utils import AzureFoundryModelInfo

        # Preserve the original model from litellm_params (includes routing prefixes like model_router/)
        # This ensures cost tracking and logging use the full model path
        original_model: str = litellm_params.get("model") or model
        if not original_model.startswith("azure_ai/"):
            # Add provider prefix if not already present
            model_response.model = f"azure_ai/{original_model}"
        else:
            model_response.model = original_model
        
        # Get base model for the parent call (strips routing prefixes for API compatibility)
        base_model: str = AzureFoundryModelInfo.get_base_model(model)
        
        return super().transform_response(
            model=base_model,
            raw_response=raw_response,
            model_response=model_response,
            logging_obj=logging_obj,
            request_data=request_data,
            messages=messages,
            optional_params=optional_params,
            litellm_params=litellm_params,
            encoding=encoding,
            api_key=api_key,
            json_mode=json_mode,
        )

    def calculate_additional_costs(
        self, model: str, prompt_tokens: int, completion_tokens: int
    ) -> Optional[dict]:
        """
        Calculate additional costs for Azure Model Router.
        
        Adds a flat infrastructure cost of $0.14 per M input tokens for using the Model Router.
        
        Args:
            model: The model name (should be a model router model)
            prompt_tokens: Number of prompt tokens
            completion_tokens: Number of completion tokens
            
        Returns:
            Dictionary with additional costs, or None if not applicable.
        """
        from litellm.llms.azure_ai.cost_calculator import (
            calculate_azure_model_router_flat_cost,
        )
        
        flat_cost = calculate_azure_model_router_flat_cost(
            model=model, prompt_tokens=prompt_tokens
        )
        
        if flat_cost > 0:
            return {"Azure Model Router Flat Cost": flat_cost}
        
        return None



"""Gray Swan guardrail configuration models."""

from typing import Dict, Optional

from pydantic import BaseModel, Field

from .base import GuardrailConfigModel


class GraySwanGuardrailConfigModelOptionalParams(BaseModel):
    """Optional parameters for the Gray Swan guardrail."""

    on_flagged_action: Optional[str] = Field(
        default="passthrough",
        description="Action when a violation is detected: 'block' rejects the call (400 error), 'monitor' logs only, 'passthrough' replaces response content with violation message (200 status).",
    )
    violation_threshold: Optional[float] = Field(
        default=0.5,
        ge=0.0,
        le=1.0,
        description="Threshold between 0 and 1 at which Gray Swan violations trigger the configured action.",
    )
    reasoning_mode: Optional[str] = Field(
        default=None,
        description="Gray Swan reasoning mode override. Accepted values: 'off', 'hybrid', 'thinking'.",
    )
    policy_id: Optional[str] = Field(
        default=None,
        description="Gray Swan policy identifier to apply during monitoring.",
    )
    categories: Optional[Dict[str, str]] = Field(
        default=None,
        description="Default Gray Swan category definitions to send with each request.",
    )
    fail_open: Optional[bool] = Field(
        default=True,
        description="If true (default), errors contacting Gray Swan are logged and the request proceeds. If false, errors propagate and block the request.",
    )
    guardrail_timeout: Optional[float] = Field(
        default=30.0,
        description="Timeout in seconds for calling the Gray Swan guardrail service.",
    )


class GraySwanGuardrailConfigModel(
    GuardrailConfigModel[GraySwanGuardrailConfigModelOptionalParams]
):
    """Configuration parameters for the Gray Swan guardrail."""

    api_key: Optional[str] = Field(
        default=None,
        description="API key for Gray Swan. Reads from the `GRAYSWAN_API_KEY` environment variable when omitted.",
    )
    api_base: Optional[str] = Field(
        default=None,
        description="Override for the Gray Swan API base URL. Defaults to https://api.grayswan.ai and can be set via `GRAYSWAN_API_BASE`.",
    )

    @staticmethod
    def ui_friendly_name() -> str:
        return "Gray Swan Guardrail"



from typing import Optional

from litellm.llms.openai.image_generation.gpt_transformation import (
    GPTImageGenerationConfig,
)
from litellm.secret_managers.main import get_secret_str


class LiteLLMProxyImageGenerationConfig(GPTImageGenerationConfig):
    """Configuration for image generation requests routed through LiteLLM Proxy."""
    def validate_environment(
        self,
        headers: dict,
        model: str,
        messages,
        optional_params: dict,
        litellm_params: dict,
        api_key: Optional[str] = None,
        api_base: Optional[str] = None,
    ) -> dict:
        api_key = api_key or get_secret_str("LITELLM_PROXY_API_KEY")
        headers.update({"Authorization": f"Bearer {api_key}"})
        return headers

    def get_complete_url(
        self,
        api_base: Optional[str],
        api_key: Optional[str],
        model: str,
        optional_params: dict,
        litellm_params: dict,
        stream: Optional[bool] = None,
    ) -> str:
        api_base = api_base or get_secret_str("LITELLM_PROXY_API_BASE")
        if api_base is None:
            raise ValueError(
                "api_base not set for LiteLLM Proxy route. Set in env via `LITELLM_PROXY_API_BASE`"
            )
        api_base = api_base.rstrip("/")
        return f"{api_base}/images/generations"



from datetime import datetime
from typing import TYPE_CHECKING, Any, Dict, List, Literal, Optional, Union

from pydantic import BaseModel, PrivateAttr
from typing_extensions import Required, TypedDict

from litellm.types.llms.base import LiteLLMPydanticObjectBase

if TYPE_CHECKING:
    from a2a.types import SendMessageResponse


# AgentProvider
class AgentProvider(TypedDict, total=False):
    """Represents the service provider of an agent."""

    organization: str  # required
    url: str  # required


# AgentExtension
class AgentExtension(TypedDict, total=False):
    """A declaration of a protocol extension supported by an Agent."""

    uri: str  # required
    description: Optional[str]
    required: Optional[bool]
    params: Optional[Dict[str, Any]]


# AgentCapabilities
class AgentCapabilities(TypedDict, total=False):
    """Defines optional capabilities supported by an agent."""

    streaming: Optional[bool]
    pushNotifications: Optional[bool]
    stateTransitionHistory: Optional[bool]
    extensions: Optional[List[AgentExtension]]


# SecurityScheme types
class SecuritySchemeBase(TypedDict, total=False):
    """Base properties shared by all security scheme objects."""

    description: Optional[str]


class APIKeySecurityScheme(SecuritySchemeBase):
    """Defines a security scheme using an API key."""

    type: Literal["apiKey"]
    in_: Literal["query", "header", "cookie"]  # using in_ to avoid Python keyword
    name: str


class HTTPAuthSecurityScheme(SecuritySchemeBase):
    """Defines a security scheme using HTTP authentication."""

    type: Literal["http"]
    scheme: str
    bearerFormat: Optional[str]


class MutualTLSSecurityScheme(SecuritySchemeBase):
    """Defines a security scheme using mTLS authentication."""

    type: Literal["mutualTLS"]


class OAuthFlows(TypedDict, total=False):
    """Defines the configuration for the supported OAuth 2.0 flows."""

    authorizationCode: Optional[Dict[str, Any]]
    clientCredentials: Optional[Dict[str, Any]]
    implicit: Optional[Dict[str, Any]]
    password: Optional[Dict[str, Any]]


class OAuth2SecurityScheme(SecuritySchemeBase):
    """Defines a security scheme using OAuth 2.0."""

    type: Literal["oauth2"]
    flows: OAuthFlows
    oauth2MetadataUrl: Optional[str]


class OpenIdConnectSecurityScheme(SecuritySchemeBase):
    """Defines a security scheme using OpenID Connect."""

    type: Literal["openIdConnect"]
    openIdConnectUrl: str


# Union of all security schemes
SecurityScheme = Union[
    APIKeySecurityScheme,
    HTTPAuthSecurityScheme,
    OAuth2SecurityScheme,
    OpenIdConnectSecurityScheme,
    MutualTLSSecurityScheme,
]


# AgentSkill
class AgentSkill(TypedDict, total=False):
    """Represents a distinct capability or function that an agent can perform."""

    id: str  # required
    name: str  # required
    description: str  # required
    tags: List[str]  # required
    examples: Optional[List[str]]
    inputModes: Optional[List[str]]
    outputModes: Optional[List[str]]
    security: Optional[List[Dict[str, List[str]]]]


# AgentInterface
class AgentInterface(TypedDict, total=False):
    """Declares a combination of a target URL and a transport protocol."""

    url: str  # required
    transport: str  # required (TransportProtocol | string)


# AgentCardSignature
class AgentCardSignature(TypedDict, total=False):
    """Represents a JWS signature of an AgentCard."""

    protected: str  # required
    signature: str  # required
    header: Optional[Dict[str, Any]]


# AgentCard
class AgentCard(TypedDict, total=False):
    """
    The AgentCard is a self-describing manifest for an agent.
    It provides essential metadata including the agent's identity, capabilities,
    skills, supported communication methods, and security requirements.
    """

    # Required fields
    protocolVersion: str
    name: str
    description: str
    url: str
    version: str
    capabilities: AgentCapabilities
    defaultInputModes: List[str]
    defaultOutputModes: List[str]
    skills: List[AgentSkill]

    # Optional fields
    preferredTransport: Optional[str]
    additionalInterfaces: Optional[List[AgentInterface]]
    iconUrl: Optional[str]
    provider: Optional[AgentProvider]
    documentationUrl: Optional[str]
    securitySchemes: Optional[Dict[str, SecurityScheme]]
    security: Optional[List[Dict[str, List[str]]]]
    supportsAuthenticatedExtendedCard: Optional[bool]
    signatures: Optional[List[AgentCardSignature]]


class AugmentedAgentCard(AgentCard):
    is_public: bool


class AgentConfig(TypedDict, total=False):
    agent_name: Required[str]
    agent_card_params: Required[AgentCard]
    litellm_params: Dict[str, Any]  # allow for any future litellm params


class PatchAgentRequest(TypedDict, total=False):
    agent_name: str
    agent_card_params: AgentCard
    litellm_params: Dict[str, Any]


# Request/Response models for CRUD endpoints


class AgentResponse(BaseModel):
    agent_id: str
    agent_name: str
    litellm_params: Optional[Dict[str, Any]] = None
    agent_card_params: Dict[str, Any]
    created_at: Optional[datetime] = None
    updated_at: Optional[datetime] = None
    created_by: Optional[str] = None
    updated_by: Optional[str] = None


class ListAgentsResponse(BaseModel):
    agents: List[AgentResponse]


class AgentMakePublicResponse(BaseModel):
    message: str
    public_agent_groups: List[str]
    updated_by: str


class MakeAgentsPublicRequest(BaseModel):
    agent_ids: List[str]


class LiteLLMSendMessageResponse(LiteLLMPydanticObjectBase):
    """
    LiteLLM wrapper for A2A SendMessageResponse.

    Wraps the a2a SDK's SendMessageResponse with LiteLLM's _hidden_params
    for cost tracking and logging integration.
    """

    # A2A response fields
    id: str
    jsonrpc: str = "2.0"
    result: Optional[Dict[str, Any]] = None
    error: Optional[Dict[str, Any]] = None

    # LiteLLM usage tracking
    usage: Optional[Dict[str, Any]] = None

    model_config = {"extra": "allow"}

    # LiteLLM private attributes for logging/cost tracking
    _hidden_params: dict = PrivateAttr(default_factory=dict)

    @classmethod
    def from_a2a_response(
        cls, response: "SendMessageResponse"
    ) -> "LiteLLMSendMessageResponse":
        """
        Create a LiteLLMSendMessageResponse from an a2a SDK SendMessageResponse.

        Args:
            response: The a2a SDK SendMessageResponse

        Returns:
            LiteLLMSendMessageResponse with _hidden_params support
        """
        # Convert the a2a response to a dict
        response_dict = response.model_dump(mode="json", exclude_none=True)

        return cls(**response_dict)

    @classmethod
    def from_dict(cls, response_dict: Dict[str, Any]) -> "LiteLLMSendMessageResponse":
        """
        Create a LiteLLMSendMessageResponse from a dict.

        Args:
            response_dict: Dict with A2A response structure

        Returns:
            LiteLLMSendMessageResponse with _hidden_params support
        """
        return cls(**response_dict)



"""
RunwayML Text-to-Speech transformation

Maps OpenAI TTS spec to RunwayML Text-to-Speech API
"""
import asyncio
import time
from typing import TYPE_CHECKING, Any, Coroutine, Dict, Optional, Tuple, Union

import httpx

import litellm
from litellm._logging import verbose_logger
from litellm.constants import (
    RUNWAYML_DEFAULT_API_VERSION,
    RUNWAYML_POLLING_TIMEOUT,
)
from litellm.llms.base_llm.text_to_speech.transformation import (
    BaseTextToSpeechConfig,
    TextToSpeechRequestData,
)
from litellm.secret_managers.main import get_secret_str

if TYPE_CHECKING:
    from litellm.litellm_core_utils.litellm_logging import Logging as LiteLLMLoggingObj
    from litellm.types.llms.openai import HttpxBinaryResponseContent
else:
    LiteLLMLoggingObj = Any
    HttpxBinaryResponseContent = Any


class RunwayMLTextToSpeechConfig(BaseTextToSpeechConfig):
    """
    Configuration for RunwayML Text-to-Speech
    
    Reference: https://api.dev.runwayml.com/v1/text_to_speech
    """
    
    DEFAULT_BASE_URL: str = "https://api.dev.runwayml.com"
    TTS_ENDPOINT_PATH: str = "v1/text_to_speech"
    DEFAULT_MODEL: str = "eleven_multilingual_v2"
    DEFAULT_VOICE_TYPE: str = "runway-preset"
    DEFAULT_VOICE_PRESET_ID: str = "Bernard"
    
    # Voice mappings from OpenAI voices to RunwayML preset IDs
    # OpenAI voices mapped to similar-sounding RunwayML voices
    VOICE_MAPPINGS = {
        "alloy": "Maya",      # Neutral, balanced female voice
        "echo": "James",      # Male voice
        "fable": "Bernard",   # Warm, storytelling voice
        "onyx": "Vincent",    # Deep male voice
        "nova": "Serene",     # Warm, expressive female voice
        "shimmer": "Ella",    # Clear, friendly female voice
    }

    def dispatch_text_to_speech(
        self,
        model: str,
        input: str,
        voice: Optional[Union[str, Dict]],
        optional_params: Dict,
        litellm_params_dict: Dict,
        logging_obj: "LiteLLMLoggingObj",
        timeout: Union[float, httpx.Timeout],
        extra_headers: Optional[Dict[str, Any]],
        base_llm_http_handler: Any,
        aspeech: bool,
        api_base: Optional[str],
        api_key: Optional[str],
        **kwargs: Any,
    ) -> Union[
        "HttpxBinaryResponseContent",
        Coroutine[Any, Any, "HttpxBinaryResponseContent"],
    ]:
        """
        Dispatch method to handle RunwayML TTS requests
        
        This method encapsulates RunwayML-specific credential resolution and parameter handling
        
        Args:
            base_llm_http_handler: The BaseLLMHTTPHandler instance from main.py
        """
        # Resolve api_base from multiple sources
        api_base = (
            api_base
            or litellm_params_dict.get("api_base")
            or litellm.api_base
            or get_secret_str("RUNWAYML_API_BASE")
            or self.DEFAULT_BASE_URL
        )
        
        # Resolve api_key from multiple sources
        api_key = (
            api_key
            or litellm_params_dict.get("api_key")
            or litellm.api_key
            or get_secret_str("RUNWAYML_API_SECRET")
            or get_secret_str("RUNWAYML_API_KEY")
        )
        
        # Convert voice to appropriate format
        voice_param: Optional[Union[str, Dict]] = voice
        if isinstance(voice, str):
            # Keep as string, will be processed in map_openai_params
            voice_param = voice
        elif isinstance(voice, dict):
            # Already in dict format, pass through
            voice_param = voice
        
        litellm_params_dict.update({
            "api_key": api_key,
            "api_base": api_base,
        })
        
        # Call the text_to_speech_handler
        response = base_llm_http_handler.text_to_speech_handler(
            model=model,
            input=input,
            voice=voice_param,
            text_to_speech_provider_config=self,
            text_to_speech_optional_params=optional_params,
            custom_llm_provider="runwayml",
            litellm_params=litellm_params_dict,
            logging_obj=logging_obj,
            timeout=timeout,
            extra_headers=extra_headers,
            client=None,
            _is_async=aspeech,
        )
        
        return response

    def get_supported_openai_params(self, model: str) -> list:
        """
        RunwayML TTS supports these OpenAI parameters
        """
        return ["voice"]

    def map_openai_params(
        self,
        model: str,
        optional_params: Dict,
        voice: Optional[Union[str, Dict]] = None,
        drop_params: bool = False,
        kwargs: Dict = {},
    ) -> Tuple[Optional[str], Dict]:
        """
        Map OpenAI parameters to RunwayML TTS parameters
        
        Returns:
            Tuple of (mapped_voice_string, mapped_params)
            
        Note: Since RunwayML requires voice as a dict, we store it in
        mapped_params["runwayml_voice"] and return None for the voice string.
        """
        mapped_params = {}
        
        # Map voice parameter to RunwayML format dict
        voice_dict: Optional[Dict] = None
        if isinstance(voice, str):
            # Check if it's an OpenAI voice name that needs mapping
            if voice in self.VOICE_MAPPINGS:
                preset_id = self.VOICE_MAPPINGS[voice]
                voice_dict = {
                    "type": self.DEFAULT_VOICE_TYPE,
                    "presetId": preset_id,
                }
            else:
                # Assume it's a RunwayML preset ID
                voice_dict = {
                    "type": self.DEFAULT_VOICE_TYPE,
                    "presetId": voice,
                }
        elif isinstance(voice, dict):
            # Already in RunwayML format, use as-is
            voice_dict = voice
        
        # Store the voice dict in optional_params for later use
        if voice_dict is not None:
            mapped_params["runwayml_voice"] = voice_dict
        
        # No other OpenAI params are currently supported by RunwayML TTS
        # (response_format, speed, etc. are not supported)
        
        # Return None for voice string since RunwayML uses dict format
        return None, mapped_params

    def validate_environment(
        self,
        headers: dict,
        model: str,
        api_key: Optional[str] = None,
        api_base: Optional[str] = None,
    ) -> dict:
        """
        Validate RunwayML environment and set up authentication headers
        """
        validated_headers = headers.copy()
        
        final_api_key = (
            api_key 
            or get_secret_str("RUNWAYML_API_SECRET") 
            or get_secret_str("RUNWAYML_API_KEY")
        )
        
        if not final_api_key:
            raise ValueError("RUNWAYML_API_SECRET or RUNWAYML_API_KEY is not set")
        
        validated_headers["Authorization"] = f"Bearer {final_api_key}"
        validated_headers["X-Runway-Version"] = RUNWAYML_DEFAULT_API_VERSION
        validated_headers["Content-Type"] = "application/json"
        
        return validated_headers

    def get_complete_url(
        self,
        model: str,
        api_base: Optional[str],
        litellm_params: dict,
    ) -> str:
        """
        Get the complete URL for RunwayML TTS request
        """
        complete_url = (
            api_base 
            or get_secret_str("RUNWAYML_API_BASE") 
            or self.DEFAULT_BASE_URL
        )
        
        complete_url = complete_url.rstrip("/")
        return f"{complete_url}/{self.TTS_ENDPOINT_PATH}"

    @staticmethod
    def _check_timeout(start_time: float, timeout_secs: float) -> None:
        """
        Check if operation has timed out.
        
        Args:
            start_time: Start time of the operation
            timeout_secs: Timeout duration in seconds
            
        Raises:
            TimeoutError: If operation has exceeded timeout
        """
        if time.time() - start_time > timeout_secs:
            raise TimeoutError(
                f"RunwayML TTS task polling timed out after {timeout_secs} seconds"
            )

    @staticmethod
    def _check_task_status(response_data: Dict[str, Any]) -> str:
        """
        Check RunwayML task status from response.
        
        RunwayML statuses: PENDING, RUNNING, SUCCEEDED, FAILED, CANCELLED, THROTTLED
        
        Args:
            response_data: JSON response from RunwayML task endpoint
            
        Returns:
            Normalized status string: "running", "succeeded", or raises on failure
            
        Raises:
            ValueError: If task failed or status is unknown
        """
        status = response_data.get("status", "").upper()
        
        verbose_logger.debug(f"RunwayML TTS task status: {status}")
        
        if status == "SUCCEEDED":
            return "succeeded"
        elif status == "FAILED":
            failure_reason = response_data.get("failure", "Unknown error")
            failure_code = response_data.get("failureCode", "unknown")
            raise ValueError(
                f"RunwayML TTS failed: {failure_reason} (code: {failure_code})"
            )
        elif status == "CANCELLED":
            raise ValueError("RunwayML TTS was cancelled")
        elif status in ["PENDING", "RUNNING", "THROTTLED"]:
            return "running"
        else:
            raise ValueError(f"Unknown RunwayML task status: {status}")

    def _poll_task_sync(
        self,
        task_id: str,
        api_base: str,
        headers: Dict[str, str],
        timeout_secs: float = 600,
    ) -> httpx.Response:
        """
        Poll RunwayML task until completion (sync).
        
        RunwayML POST returns immediately with a task that has status PENDING/RUNNING.
        We need to poll GET /v1/tasks/{task_id} until status is SUCCEEDED or FAILED.
        
        Args:
            task_id: The task ID to poll
            api_base: Base URL for RunwayML API
            headers: Request headers (including auth)
            timeout_secs: Total timeout in seconds (default: 600s = 10 minutes)
            
        Returns:
            Final response with completed task
        """
        from litellm.llms.custom_httpx.http_handler import _get_httpx_client

        client = _get_httpx_client()
        start_time = time.time()
        
        # Build task status URL
        api_base = api_base.rstrip("/")
        task_url = f"{api_base}/v1/tasks/{task_id}"
        
        verbose_logger.debug(f"Polling RunwayML TTS task: {task_url}")
        
        while True:
            self._check_timeout(start_time=start_time, timeout_secs=timeout_secs)
            
            # Poll the task status
            response = client.get(url=task_url, headers=headers)
            response.raise_for_status()
            
            response_data = response.json()
            
            # Check task status
            status = self._check_task_status(response_data=response_data)
            
            if status == "succeeded":
                return response
            elif status == "running":
                # Wait before polling again (RunwayML recommends 1-2 second intervals)
                time.sleep(2)

    async def _poll_task_async(
        self,
        task_id: str,
        api_base: str,
        headers: Dict[str, str],
        timeout_secs: float = 600,
    ) -> httpx.Response:
        """
        Poll RunwayML task until completion (async).
        
        Args:
            task_id: The task ID to poll
            api_base: Base URL for RunwayML API
            headers: Request headers (including auth)
            timeout_secs: Total timeout in seconds (default: 600s = 10 minutes)
            
        Returns:
            Final response with completed task
        """
        from litellm.llms.custom_httpx.http_handler import get_async_httpx_client

        client = get_async_httpx_client(llm_provider=litellm.LlmProviders.RUNWAYML)
        start_time = time.time()
        
        # Build task status URL
        api_base = api_base.rstrip("/")
        task_url = f"{api_base}/v1/tasks/{task_id}"
        
        verbose_logger.debug(f"Polling RunwayML TTS task (async): {task_url}")
        
        while True:
            self._check_timeout(start_time=start_time, timeout_secs=timeout_secs)
            
            # Poll the task status
            response = await client.get(url=task_url, headers=headers)
            response.raise_for_status()
            
            response_data = response.json()
            
            # Check task status
            status = self._check_task_status(response_data=response_data)
            
            if status == "succeeded":
                return response
            elif status == "running":
                # Wait before polling again (RunwayML recommends 1-2 second intervals)
                await asyncio.sleep(2)

    def transform_text_to_speech_request(
        self,
        model: str,
        input: str,
        voice: Optional[Union[str, Dict]],
        optional_params: Dict,
        litellm_params: Dict,
        headers: dict,
    ) -> TextToSpeechRequestData:
        """
        Transform OpenAI TTS request to RunwayML TTS format
        
        RunwayML expects:
        - model: The model to use (e.g., 'eleven_multilingual_v2')
        - promptText: The text to convert to speech
        - voice: Voice configuration object
          {
            "type": "runway-preset",
            "presetId": "Bernard"
          }
        
        Returns:
            TextToSpeechRequestData: Contains JSON body and headers
        """
        # Get voice from optional_params (mapped in map_openai_params)
        runwayml_voice = optional_params.get("runwayml_voice")
        if runwayml_voice is None:
            # Use default voice if not provided
            runwayml_voice = {
                "type": self.DEFAULT_VOICE_TYPE,
                "presetId": self.DEFAULT_VOICE_PRESET_ID,
            }
        
        # Build request body
        request_body = {
            "model": model or self.DEFAULT_MODEL,
            "promptText": input,
            "voice": runwayml_voice,
        }
        
        # Add any other optional parameters (except runwayml_voice which we already used)
        for k, v in optional_params.items():
            if k not in request_body and k != "runwayml_voice":
                request_body[k] = v
        
        return {
            "dict_body": request_body,
            "headers": headers,
        }

    def transform_text_to_speech_response(
        self,
        model: str,
        raw_response: httpx.Response,
        logging_obj: "LiteLLMLoggingObj",
    ) -> "HttpxBinaryResponseContent":
        """
        Transform RunwayML TTS response to standard format
        
        RunwayML returns a task immediately with status PENDING/RUNNING.
        We need to poll the task until it completes, then download the audio.
        
        Initial response:
        {
            "id": "task_123...",
            "status": "PENDING" | "RUNNING",
            "createdAt": "2025-11-13T..."
        }
        
        After polling:
        {
            "id": "task_123...",
            "status": "SUCCEEDED",
            "output": ["https://storage.googleapis.com/.../audio.mp3"],
            "completedAt": "2025-11-13T..."
        }
        """
        from litellm.types.llms.openai import HttpxBinaryResponseContent

        try:
            response_data = raw_response.json()
        except Exception as e:
            raise self.get_error_class(
                error_message=f"Error parsing RunwayML TTS response: {e}",
                status_code=raw_response.status_code,
                headers=dict(raw_response.headers),
            )
        
        verbose_logger.debug("RunwayML TTS starting polling...")
        
        # Get task ID
        task_id = response_data.get("id")
        if not task_id:
            raise ValueError("RunwayML TTS response missing task ID")
        
        # Get headers for polling (need auth)
        poll_headers = {
            "Authorization": raw_response.request.headers.get("Authorization", ""),
            "X-Runway-Version": raw_response.request.headers.get(
                "X-Runway-Version", RUNWAYML_DEFAULT_API_VERSION
            ),
        }
        
        # Poll until task completes
        polled_response = self._poll_task_sync(
            task_id=task_id,
            api_base=self.DEFAULT_BASE_URL,
            headers=poll_headers,
            timeout_secs=RUNWAYML_POLLING_TIMEOUT,
        )
        
        # Get the completed task data
        task_data = polled_response.json()
        
        verbose_logger.debug("RunwayML TTS polling complete, downloading audio")
        
        # Get audio URL from output
        output = task_data.get("output", [])
        if not output or not isinstance(output, list) or len(output) == 0:
            raise ValueError("RunwayML TTS response missing audio URL in output")
        
        audio_url = output[0]
        if not isinstance(audio_url, str):
            raise ValueError(f"RunwayML TTS audio URL is not a string: {audio_url}")
        
        # Download the audio file
        from litellm.llms.custom_httpx.http_handler import _get_httpx_client

        client = _get_httpx_client()
        audio_response = client.get(url=audio_url)
        audio_response.raise_for_status()
        
        verbose_logger.debug("RunwayML TTS audio downloaded successfully")
        
        # Return the audio data wrapped in HttpxBinaryResponseContent
        return HttpxBinaryResponseContent(audio_response)

    async def async_transform_text_to_speech_response(
        self,
        model: str,
        raw_response: httpx.Response,
        logging_obj: "LiteLLMLoggingObj",
    ) -> "HttpxBinaryResponseContent":
        """
        Async transform RunwayML TTS response to standard format
        
        Same as sync version but uses async polling and download
        """
        from litellm.types.llms.openai import HttpxBinaryResponseContent

        try:
            response_data = raw_response.json()
        except Exception as e:
            raise self.get_error_class(
                error_message=f"Error parsing RunwayML TTS response: {e}",
                status_code=raw_response.status_code,
                headers=dict(raw_response.headers),
            )
        
        verbose_logger.debug("RunwayML TTS starting polling (async)...")
        
        # Get task ID
        task_id = response_data.get("id")
        if not task_id:
            raise ValueError("RunwayML TTS response missing task ID")
        
        # Get headers for polling (need auth)
        poll_headers = {
            "Authorization": raw_response.request.headers.get("Authorization", ""),
            "X-Runway-Version": raw_response.request.headers.get(
                "X-Runway-Version", RUNWAYML_DEFAULT_API_VERSION
            ),
        }
        
        # Poll until task completes (async)
        polled_response = await self._poll_task_async(
            task_id=task_id,
            api_base=self.DEFAULT_BASE_URL,
            headers=poll_headers,
            timeout_secs=RUNWAYML_POLLING_TIMEOUT,
        )
        
        # Get the completed task data
        task_data = polled_response.json()
        
        verbose_logger.debug("RunwayML TTS polling complete (async), downloading audio")
        
        # Get audio URL from output
        output = task_data.get("output", [])
        if not output or not isinstance(output, list) or len(output) == 0:
            raise ValueError("RunwayML TTS response missing audio URL in output")
        
        audio_url = output[0]
        if not isinstance(audio_url, str):
            raise ValueError(f"RunwayML TTS audio URL is not a string: {audio_url}")
        
        # Download the audio file (async)
        from litellm.llms.custom_httpx.http_handler import get_async_httpx_client

        client = get_async_httpx_client(llm_provider=litellm.LlmProviders.RUNWAYML)
        audio_response = await client.get(url=audio_url)
        audio_response.raise_for_status()
        
        verbose_logger.debug("RunwayML TTS audio downloaded successfully (async)")
        
        # Return the audio data wrapped in HttpxBinaryResponseContent
        return HttpxBinaryResponseContent(audio_response)




"""
This file contains the calling Azure OpenAI's `/openai/realtime` endpoint.

This requires websockets, and is currently only supported on LiteLLM Proxy.
"""

from typing import Any, Optional, cast

from litellm.constants import REALTIME_WEBSOCKET_MAX_MESSAGE_SIZE_BYTES

from ....litellm_core_utils.litellm_logging import Logging as LiteLLMLogging
from ....litellm_core_utils.realtime_streaming import RealTimeStreaming
from ....llms.custom_httpx.http_handler import get_shared_realtime_ssl_context
from ..azure import AzureChatCompletion
from litellm._logging import verbose_proxy_logger

# BACKEND_WS_URL = "ws://localhost:8080/v1/realtime?model=gpt-4o-realtime-preview-2024-10-01"


async def forward_messages(client_ws: Any, backend_ws: Any):
    import websockets

    try:
        while True:
            message = await backend_ws.recv()
            await client_ws.send_text(message)
    except websockets.exceptions.ConnectionClosed:  # type: ignore
        pass


class AzureOpenAIRealtime(AzureChatCompletion):
    def _construct_url(
        self,
        api_base: str,
        model: str,
        api_version: str,
        realtime_protocol: Optional[str] = None,
    ) -> str:
        """
        Construct Azure realtime WebSocket URL.

        Args:
            api_base: Azure API base URL (will be converted from https:// to wss://)
            model: Model deployment name
            api_version: Azure API version
            realtime_protocol: Protocol version to use:
                - "GA" or "v1": Uses /openai/v1/realtime (GA path)
                - "beta" or None: Uses /openai/realtime (beta path, default)

        Returns:
            WebSocket URL string

        Examples:
            beta/default: "wss://.../openai/realtime?api-version=2024-10-01-preview&deployment=gpt-4o-realtime-preview"
            GA/v1:        "wss://.../openai/v1/realtime?model=gpt-realtime-deployment"
        """
        api_base = api_base.replace("https://", "wss://")

        # Determine path based on realtime_protocol
        if realtime_protocol in ("GA", "v1"):
            path = "/openai/v1/realtime" 
            return f"{api_base}{path}?model={model}"
        else:
            # Default to beta path for backwards compatibility
            path = "/openai/realtime"
            return f"{api_base}{path}?api-version={api_version}&deployment={model}"

    async def async_realtime(
        self,
        model: str,
        websocket: Any,
        logging_obj: LiteLLMLogging,
        api_base: Optional[str] = None,
        api_key: Optional[str] = None,
        api_version: Optional[str] = None,
        azure_ad_token: Optional[str] = None,
        client: Optional[Any] = None,
        timeout: Optional[float] = None,
        realtime_protocol: Optional[str] = None,
    ):
        import websockets
        from websockets.asyncio.client import ClientConnection

        if api_base is None:
            raise ValueError("api_base is required for Azure OpenAI calls")
        if api_version is None:
            raise ValueError("api_version is required for Azure OpenAI calls")

        url = self._construct_url(
            api_base, model, api_version, realtime_protocol=realtime_protocol
        )

        try:
            ssl_context = get_shared_realtime_ssl_context()
            async with websockets.connect(  # type: ignore
                url,
                additional_headers={
                    "api-key": api_key,  # type: ignore
                },
                max_size=REALTIME_WEBSOCKET_MAX_MESSAGE_SIZE_BYTES,
                ssl=ssl_context,
            ) as backend_ws:
                realtime_streaming = RealTimeStreaming(
                    websocket, cast(ClientConnection, backend_ws), logging_obj
                )
                await realtime_streaming.bidirectional_forward()

        except websockets.exceptions.InvalidStatusCode as e:  # type: ignore
            await websocket.close(code=e.status_code, reason=str(e))
        except Exception:
            verbose_proxy_logger.exception("Error in AzureOpenAIRealtime.async_realtime")
            pass
